# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# This is based on the specifications of:
# X. Li, K. Tang, M. Omidvar, Z. Yang and K. Qin,
# "Benchmark Functions for the CEC’2013 Special Session and Competition on Large Scale Global Optimization",
# Technical Report, Evolutionary Computation and Machine Learning Group, RMIT University, Australia, 2013.
# and the corresponding source code.

from typing import List, Callable, Dict, Type, Optional
from pathlib import Path
import re
import numpy as np
from nevergrad.common.decorators import Registry
from nevergrad.functions import ExperimentFunction
import nevergrad as ng
from . import _core


registry = Registry[Type["LsgoFunction"]]()


def read_data(name: str) -> Dict[str, np.ndarray]:
    filepaths = (Path(__file__).parent / "cdatafiles").glob(name + "-*.txt")
    all_data: Dict[str, np.ndarray] = {}
    pattern = re.compile(name + "-(?P<tag>.+?).txt")
    for filepath in filepaths:
        match = pattern.search(filepath.name)
        assert match is not None
        tag = match.group("tag")
        all_data[tag] = np.loadtxt(str(filepath), delimiter=",", dtype=int if tag in ["s", "p"] else float)
    if "p" in all_data:
        all_data["p"] -= 1  # indexing starts at 0 in Python
    return all_data


class FunctionChunk:
    """Computes a sequence of transforms and applies a loss on it.
    This can be left multiplied by a scalar to add a weight.
    """

    def __init__(
        self, transforms: List[Callable[[np.ndarray], np.ndarray]], loss: Callable[[np.ndarray], float]
    ) -> None:
        self.loss = loss
        self.transforms = transforms
        self._scalar = 1.0

    def _apply_transforms(self, x: np.ndarray) -> np.ndarray:
        if not self.transforms:
            return x
        y = self.transforms[0](x)
        for transf in self.transforms[1:]:
            y = transf(y)
        return y

    def __call__(self, x: np.ndarray) -> float:
        return self._scalar * self.loss(self._apply_transforms(x))

    def __rmul__(self, scalar: float) -> "FunctionChunk":
        self._scalar *= scalar
        return self

    def __repr__(self) -> str:
        return f"{self._scalar}*FunctionChunck({self.transforms}, {self.loss}"


class LsgoFunction:
    """Base function used in the LSGO testbed.
    It is in charge of computing the translation, and summing all function chunks.
    The "optimum" attribute holds the zero of the function. It is usually xopt, but is xopt + 1 in one case (F12),
    and unknown in one other (F14).
    """

    bounds = (-100, 100)
    tags = ["unimodal", "separable", "shifted", "irregular"]
    conditionning = -1.0

    def __init__(self, xopt: np.ndarray, functions: List[FunctionChunk]) -> None:
        self.xopt = xopt
        self.optimum: Optional[np.ndarray] = xopt  # default to xopt, but requires changing for rosenbroc
        self.functions = functions

    def __call__(self, x: np.ndarray) -> float:
        transformed = x - self.xopt
        return sum(f(transformed) for f in self.functions)

    @property
    def dimension(self) -> int:
        """Dimension of the function space"""
        return self.xopt.size

    def instrumented(self, transform: str = "bouncing") -> ExperimentFunction:
        """Returns an instrumented function, taking the bounds into account by composing it with
        a bounding transform. Instrumentated functions are necessary
        for nevergrad benchmarking

        Parameter
        ---------
        transform: str
            "bouncing", "arctan", "tanh" or "clipping"
        """
        param = ng.p.Array(shape=(self.dimension,)).set_bounds(
            self.bounds[0], self.bounds[1], method=transform
        )
        return ExperimentFunction(self, param.set_name(transform))


@registry.register
class ShiftedElliptic(LsgoFunction):

    number = 1
    bounds = (-100, 100)
    conditionning = 1e6
    tags = ["unimodal", "separable", "shifted", "irregular"]

    def __init__(self, xopt: np.ndarray) -> None:
        transforms: List[Callable[[np.ndarray], np.ndarray]] = [_core.irregularity]
        super().__init__(xopt, [FunctionChunk(transforms, _core.Elliptic(xopt.size))])


@registry.register
class ShiftedRastrigin(LsgoFunction):

    number = 2
    bounds = (-5, 5)
    conditionning = 10.0
    tags = ["multimodal", "separable", "shifted", "irregular"]

    def __init__(self, xopt: np.ndarray) -> None:
        transforms: List[Callable[[np.ndarray], np.ndarray]] = [
            _core.irregularity,
            _core.Asymmetry(0.2),
            _core.Illconditionning(10.0),
        ]
        super().__init__(xopt, [FunctionChunk(transforms, _core.rastrigin)])


@registry.register
class ShiftedAckley(LsgoFunction):

    number = 3
    bounds = (-32, 32)
    conditionning = 10.0
    tags = ["multimodal", "separable", "shifted", "irregular"]

    def __init__(self, xopt: np.ndarray) -> None:
        transforms: List[Callable[[np.ndarray], np.ndarray]] = [
            _core.irregularity,
            _core.Asymmetry(0.2),
            _core.Illconditionning(10.0),
        ]
        super().__init__(xopt, [FunctionChunk(transforms, _core.ackley)])


# %% PARTIALLY SEPARABLE 1 %% #


class _MultiPartFunction(LsgoFunction):
    """Base class for most multi-part function, overlapping or not."""

    number = -1
    bounds = (-100, 100)
    tags: List[str] = []
    conditionning = -1.0
    overlap = 0

    def _make_loss(self, dimension: int, side_loss: bool) -> Callable[[np.ndarray], float]:
        raise NotImplementedError

    def _make_transforms(self, side_loss: bool) -> List[Callable[[np.ndarray], np.ndarray]]:
        raise NotImplementedError

    def __init__(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        xopt: np.ndarray,
        p: np.ndarray,
        s: np.ndarray,
        w: np.ndarray,
        R25: np.ndarray,
        R50: np.ndarray,
        R100: np.ndarray,
    ) -> None:
        remaining = 1000 - np.sum(s)
        assert remaining in [0, 700]
        assert xopt.size == p.size
        indexings = _core.Indexing.from_split(
            p, s.tolist() + ([remaining] if remaining else []), overlap=self.overlap
        )
        rotations = {num: _core.Rotation(rot) for num, rot in [(25, R25), (50, R50), (100, R100)]}
        assert w.size == len(indexings) - (remaining == 700)
        functions: List[FunctionChunk] = []
        transforms: List[Callable[[np.ndarray], np.ndarray]] = []
        for coeff, indexing in zip(w.tolist(), indexings):
            transforms = [indexing, rotations[indexing.outdim]] + self._make_transforms(side_loss=False)  # type: ignore
            functions.append(
                coeff * FunctionChunk(transforms, self._make_loss(indexing.outdim, side_loss=False))
            )
        if remaining:
            transforms = self._make_transforms(side_loss=True)
            loss = self._make_loss(indexings[-1].outdim, side_loss=True)
            functions.append(FunctionChunk([indexings[-1]] + transforms, loss))  # type: ignore
        super().__init__(xopt, functions)


@registry.register
class PartiallySeparableElliptic(_MultiPartFunction):

    number = 4
    bounds = (-100, 100)
    tags = ["unimodal", "partially separable", "shifted", "irregularities"]
    conditionning = 1e6

    def _make_loss(self, dimension: int, side_loss: bool) -> Callable[[np.ndarray], float]:
        return _core.Elliptic(dimension)

    def _make_transforms(self, side_loss: bool) -> List[Callable[[np.ndarray], np.ndarray]]:
        return [_core.irregularity]


@registry.register
class PartiallySeparableRastrigin(_MultiPartFunction):

    number = 5
    bounds = (-5, 5)
    tags = ["multimodal", "partially separable", "shifted", "irregularities"]
    conditionning = 10.0

    def _make_loss(self, dimension: int, side_loss: bool) -> Callable[[np.ndarray], float]:
        return _core.rastrigin

    def _make_transforms(self, side_loss: bool) -> List[Callable[[np.ndarray], np.ndarray]]:
        return [_core.irregularity, _core.Asymmetry(0.2), _core.Illconditionning(10.0)]


@registry.register
class PartiallySeparableAckley(_MultiPartFunction):

    number = 6
    bounds = (-32, 32)
    tags = ["multimodal", "partially separable", "shifted", "irregularities"]
    conditionning = 10.0

    def _make_loss(self, dimension: int, side_loss: bool) -> Callable[[np.ndarray], float]:
        return _core.ackley

    def _make_transforms(self, side_loss: bool) -> List[Callable[[np.ndarray], np.ndarray]]:
        return [_core.irregularity, _core.Asymmetry(0.2), _core.Illconditionning(10.0)]


@registry.register
class PartiallySeparableSchwefel(_MultiPartFunction):

    number = 7
    bounds = (-100, 100)
    tags = ["multimodal", "partially separable", "shifted", "irregularities"]
    conditionning = -1.0  # not provided

    def _make_loss(self, dimension: int, side_loss: bool) -> Callable[[np.ndarray], float]:
        return _core.sphere if side_loss else _core.schwefel_1_2

    def _make_transforms(self, side_loss: bool) -> List[Callable[[np.ndarray], np.ndarray]]:
        return (
            [] if side_loss else [_core.irregularity, _core.Asymmetry(0.2)]
        )  # this reproduces implementations, not paper


# %% PARTIALLY SEPARABLE 2 %% #


@registry.register
class PartiallySeparableElliptic2(PartiallySeparableElliptic):

    number = 8
    bounds = (-100, 100)
    tags = ["unimodal", "partially separable", "shifted", "irregularities"]
    conditionning = 1e6


@registry.register
class PartiallySeparableRastrigin2(PartiallySeparableRastrigin):

    number = 9
    bounds = (-5, 5)
    tags = ["multimodal", "partially separable", "shifted", "irregularities"]
    conditionning = 10.0


@registry.register
class PartiallySeparableAckley2(PartiallySeparableAckley):

    number = 10
    bounds = (-32, 32)
    tags = ["multimodal", "partially separable", "shifted", "irregularities"]
    conditionning = 10.0


@registry.register
class PartiallySeparableSchwefel2(PartiallySeparableSchwefel):

    number = 11
    bounds = (-100, 100)
    tags = ["unimodal", "partially separable", "shifted", "irregularities"]
    conditionning = -1.0  # not provided


# %% Overlapping functions %% #


@registry.register
class ShiftedRosenbrock(LsgoFunction):

    number = 12
    bounds = (-100, 100)
    conditionning = -1.0  # not provided
    tags = ["multimodal", "separable", "shifted", "irregular"]

    def __init__(self, xopt: np.ndarray) -> None:
        super().__init__(xopt, [FunctionChunk([], _core.rosenbrock)])
        self.optimum = xopt + 1  # xopt and optimum are different


@registry.register
class OverlappingSchwefel(PartiallySeparableSchwefel):

    number = 13
    bounds = (-100, 100)
    tags = ["unimodal", "non-separable", "overlapping", "shifted", "irregularities"]
    overlap = 5


@registry.register
class ConflictingSchwefel(LsgoFunction):

    number = 14
    bounds = (-100, 100)
    tags = ["unimodal", "non-separable", "conflicting", "shifted", "irregularities"]
    conditionning = -1.0  # not provided

    def __init__(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        xopt: np.ndarray,
        p: np.ndarray,
        s: np.ndarray,
        w: np.ndarray,
        R25: np.ndarray,
        R50: np.ndarray,
        R100: np.ndarray,
    ) -> None:
        assert xopt.size == 1000
        assert p.size == 905
        indexings = _core.Indexing.from_split(p, s.tolist(), overlap=5)
        xopt_indexings = _core.Indexing.from_split(np.arange(1000), s.tolist(), overlap=0)
        rotations = {num: _core.Rotation(rot) for num, rot in [(25, R25), (50, R50), (100, R100)]}
        assert w.size == len(indexings)
        functions: List[FunctionChunk] = []
        transforms: List[Callable[[np.ndarray], np.ndarray]] = []
        for coeff, indexing, xind in zip(w.tolist(), indexings, xopt_indexings):
            transforms = [
                indexing,
                _core.Translation(xopt[xind.indices]),
                rotations[indexing.outdim],
                _core.irregularity,
                _core.Asymmetry(0.2),
            ]
            functions.append(coeff * FunctionChunk(transforms, _core.schwefel_1_2))
        super().__init__(np.zeros((905,)), functions)
        self.optimum = None  # unknown


@registry.register
class ShiftedSchwefel(LsgoFunction):

    number = 15
    bounds = (-100, 100)
    tags = ["unimodal", "fully non-separable", "shifted", "irregular"]
    conditionning = -1.0  # not provided

    def __init__(self, xopt: np.ndarray) -> None:
        transforms: List[Callable[[np.ndarray], np.ndarray]] = [_core.irregularity, _core.Asymmetry(0.2)]
        super().__init__(xopt, [FunctionChunk(transforms, _core.schwefel_1_2)])


def make_function(number: int) -> LsgoFunction:
    """Creates one of the LSGO functions.

    Parameters
    ----------
    number: int
        the number of the function, from 1 to 15 (included)

    Returns
    -------
    LsgoFunction
        A function which acts exactly as the CPP implementation of LSGO (which may deviate from matlab or the
        actual paper). It has an attribute dimension for the optimization space dimension, and bounds for the upper
        and lower bounds of this space.
    """
    if not number:
        raise ValueError("Numbering of LSGO functions starts from 1")
    num_registry = {cls.number: cls for cls in registry.values()}  # type: ignore
    return num_registry[number](**read_data(f"F{number}"))  # type: ignore



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pytest
import numpy as np
import nevergrad as ng
from nevergrad.common import testing
import nevergrad.common.typing as tp
from . import _datalayers
from . import helpers


def test_scalar_module() -> None:
    ref = ng.p.Scalar()
    x = ng.p.Scalar(10) % 4
    assert x.value == 2
    assert x.get_standardized_data(reference=ref)[0] == 10
    x.value = 1
    assert x.get_standardized_data(reference=ref)[0] == 9  # find the closest


def test_bound_module() -> None:
    ref = ng.p.Scalar()
    with pytest.raises(ng.errors.NevergradValueError):
        _datalayers.Bound(3, 8, method="arctan")(ref)
    x = _datalayers.Bound(-1, 8, method="arctan")(ref)
    x.set_standardized_data([100], reference=ref)
    np.testing.assert_almost_equal(x.value, 7.97135306)


def test_log_layer() -> None:
    ref = ng.p.Scalar()
    x = 2 ** ng.p.Scalar()
    assert x.value == 1
    x.value = 16
    assert x.get_standardized_data(reference=ref)[0] == 4  # find the closest
    assert x.value == 16


def test_add_layer() -> None:
    x = ng.p.Scalar() - 4.0
    y = 6 + x
    assert y.value == 2


def test_multiply_layer() -> None:
    ref = ng.p.Scalar()
    x = ng.p.Scalar(5) / 5
    assert x.value == 1
    x.value = 2
    assert x.get_standardized_data(reference=ref)[0] == 10


def test_power() -> None:
    ref = ng.p.Scalar()
    x = 3 / ng.p.Scalar(6)
    assert x.value == 0.5
    x.value = 0.25
    assert x.get_standardized_data(reference=ref)[0] == 12


@testing.parametrized(
    new_unary=(10 ** -ng.p.Scalar(lower=0, upper=5),),
    legacy_log=(ng.p.Log(lower=1e-10, upper=1.0, exponent=5),),
    legacy_array=(
        ng.p.Array(init=1e-5 * np.ones(100)).set_bounds(lower=1e-5, upper=1.0).set_mutation(exponent=10),
    ),
)
def test_log_sampling(log: ng.p.Data) -> None:
    values: tp.List[float] = []
    while len(values) < 100:
        new = log.sample().value
        if isinstance(new, np.ndarray):
            values.extend(new.tolist())
        else:
            values.append(new)
    proba = np.mean(np.array(values) < 0.1)
    assert 0.5 < proba < 1  # should be around 80%


def test_clipping_standardized_data() -> None:
    ref = ng.p.Scalar()
    x = _datalayers.Bound(-10, 10, method="clipping")(ref)
    x.set_standardized_data([12])
    state = x.get_standardized_data(reference=ref)
    assert state[0] == 10
    assert x.value == 10


def test_bound_estimation() -> None:
    param = (_datalayers.Bound(-10, 10)(ng.p.Scalar()) + 3) * 5
    assert param.bounds == (-35, 65)  # type: ignore


def test_softmax_layer() -> None:
    param = ng.p.Array(shape=(4, 3))
    param.random_state.seed(12)
    param.add_layer(_datalayers.SoftmaxSampling(arity=3))
    assert param.value.tolist() == [0, 2, 0, 1]
    assert param.value.tolist() == [0, 2, 0, 1], "Different indices at the second call"
    del param.value
    assert param.value.tolist() == [0, 2, 2, 0], "Same indices after resampling"
    param.value = [0, 1, 2, 0]  # type: ignore
    assert param.value.tolist() == [0, 1, 2, 0]
    expected = np.zeros((4, 3))
    expected[[0, 1, 2, 3], [0, 1, 2, 0]] = 0.6931
    np.testing.assert_array_almost_equal(param._value, expected, decimal=4)


def test_deterministic_softmax_layer() -> None:
    param = ng.p.Array(shape=(1, 100))
    param.add_layer(_datalayers.SoftmaxSampling(arity=100, deterministic=True))
    param._value[0, 12] = 1
    assert param.value.tolist() == [12]


def test_temporary_deterinistic_softmax_layer() -> None:
    param = ng.p.Array(shape=(1, 100))
    param.add_layer(_datalayers.SoftmaxSampling(arity=100, deterministic=False))
    param._value[0, 12] = 1
    del param.value
    with helpers.deterministic_sampling(param):
        assert param.value[0] == 12
    # the behavior should be stochastic when not using the context
    different = False
    for _ in range(10000):
        del param.value
        different = param.value[0] != 12
        if different:
            break
    assert different, "At least one sampling should have been different"


def test_bounded_int_casting() -> None:
    param = _datalayers.Bound(-10.9, 10.9, method="clipping")(ng.p.Scalar())
    param.add_layer(_datalayers.Int())
    for move, val in [(2.4, 2), (0.2, 3), (42, 10), (-42, -10)]:
        param.set_standardized_data([move])
        assert param.value == val, f"Wrong value after move {move}"


def test_rand_int_casting() -> None:
    param = _datalayers.Bound(0, 1)(ng.p.Array(shape=(100, 10)) + 0.2)
    param = _datalayers.Int(deterministic=False)(param)
    total = param.value.ravel().sum()
    assert 50 < total < 500


@pytest.mark.parametrize("deg", (True, False))  # type: ignore
@pytest.mark.parametrize("bound_method", (None, "clipping", "arctan"))  # type: ignore
def test_angles(deg: bool, bound_method: tp.Any) -> None:
    span = 360 if deg else 2 * np.pi
    params = [
        _datalayers.Angles(shape=(10,), deg=deg, bound_method=bound_method) / span + 0.5 for _ in range(2)
    ]
    assert params[0].bounds == (0, 1)
    values = [np.linspace(0, 1, 10), np.linspace(1, 0, 10)]
    for param, value in zip(params, values):
        param.value = value
        assert param.value == pytest.approx(value)
    average = params[1].get_standardized_data(reference=params[0]) / 2
    params[1].set_standardized_data(average, reference=params[0])
    assert params[1].value.tolist() == pytest.approx([1, 1, 1, 0.5, 0.5, 0.5, 0.5, 1, 1, 1])



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import inspect
import itertools
import os
import platform
import typing as tp
from pathlib import Path
from unittest import SkipTest
import pytest
import numpy as np
from nevergrad.optimization import registry as optregistry
import nevergrad.functions.base as fbase
from nevergrad.functions.mlda import datasets
from nevergrad.functions import rl
from nevergrad.common import testing

# from nevergrad.common.tools import Selector
from nevergrad.common import tools
from .xpbase import Experiment
from .utils import Selector
from . import experiments
from . import optgroups


@testing.parametrized(**{name: (name, maker) for name, maker in experiments.registry.items()})
def test_experiments_registry(name: str, maker: tp.Callable[[], tp.Iterator[experiments.Experiment]]) -> None:
    if sum([ord(c) for c in name]) % 4 > 0:
        raise SkipTest("Too expensive: we randomly skip 3/4 of these tests.")

    # "mav" is not availablefor now.
    if "conformant" in name or name == "neuro_planning" or "sparse" in name:
        raise SkipTest("This is user parametric and can not be tested.")

    if "compiler" in name or "emulators" in name:
        raise SkipTest("Compiler/emulator stuff too heavy for CircleCI.")

    # Our PGAN is not well accepted by circleci.
    if "_pgan" in name and os.environ.get("CIRCLECI", False):
        raise SkipTest("Too slow in CircleCI")

    if "gym" in name and os.environ.get("CIRCLECI", False):
        raise SkipTest("Too slow in CircleCI")

    if "yawideb" in name or "_quality" in name:
        raise SkipTest("I should have a look at this test.")

    # mixsimulator is not accepted by circleci pytest.
    if "mixsimulator" in name and os.environ.get("CIRCLECI", False):
        raise SkipTest("Sigkill in CircleCI")

    # Our IQAs and our ScikitLearn are not well guaranteed on Windows.
    if all(x in name for x in ["image", "quality"]) and platform.system() == "Windows":
        raise SkipTest("Image quality not guaranteed on Windows.")

    # Basic test.
    print(f"Testing {name}")
    with tools.set_env(NEVERGRAD_PYTEST=1):
        with datasets.mocked_data():  # mock mlda data that should be downloaded
            check_maker(maker)  # this is to extract the function for reuse if other external packages need it

    # Some tests are skipped on CircleCI (but they do work well locally, if memory enough).
    if os.environ.get("CIRCLECI", False):
        if any(x in name for x in ["image", "mlda", "realworld", "adversarial_attack"]):
            raise SkipTest("Too slow in CircleCI")

    check_experiment(
        maker,
        ("mltuning" in name or "anm" in name),
        skip_seed=(name in ["rocket", "images_using_gan"])
        or any(x in name for x in ["tuning", "image_", "compiler", "anm", "olympus"]),
    )  # this is a basic test on first elements, do not fully rely on it


@pytest.fixture(scope="module")  # type: ignore
def recorder() -> tp.Generator[tp.Dict[str, tp.List[optgroups.Optim]], None, None]:
    record: tp.Dict[str, tp.List[optgroups.Optim]] = {}
    yield record
    groups = sorted(record.items())
    string = "\n\n".join(f"{x} = {repr(y)}" for x, y in groups)
    filepath = Path(__file__).with_name("optimizer_groups.txt")
    filepath.write_text(string)


# pylint: disable=redefined-outer-name
@pytest.mark.parametrize("name", optgroups.registry)  # type: ignore
def test_groups_registry(name: str, recorder: tp.Dict[str, tp.List[optgroups.Optim]]) -> None:
    maker = optgroups.registry[name]
    opts = list(maker())
    for opt in opts:
        if isinstance(opt, str):
            assert opt in optregistry, f"{opt} is not registered."
    recorder[name] = opts


def check_maker(maker: tp.Callable[[], tp.Iterator[experiments.Experiment]]) -> None:
    generators = [maker() for _ in range(2)]
    # check 1 sample
    try:
        sample = next(maker())
    except fbase.UnsupportedExperiment as e:
        raise SkipTest("Skipping because unsupported") from e
    assert isinstance(sample, experiments.Experiment)
    # check names, coherence and non-randomness
    for k, (elem1, elem2) in enumerate(itertools.zip_longest(*generators)):
        assert not elem1.is_incoherent, f"Incoherent settings should be filtered out from generator:\n{elem1}"
        try:
            assert elem1 == elem2  # much faster but lacks explicit message
        except AssertionError:
            testing.printed_assert_equal(
                elem1.get_description(),
                elem2.get_description(),
                err_msg=f"Two paths on the generator differed (see element #{k})\n"
                "Generators need to be deterministic in order to split the workload!",
            )


def check_experiment(maker: tp.Any, short: bool = False, skip_seed: bool = False) -> None:
    """Randomized check of seedability for 8 first elements
    This test does not prove the complete seedability of the generator!  (would be way too slow)
    """
    # we use "maker: Any" because signature for one or the other case (seedable or not) is way too complex, and won't help much here...
    random_seed = np.random.randint(1000)
    signature = inspect.signature(maker)
    if not signature.parameters:
        return  # not designed to be seedable
    # draw twice with same random seed_generator and once with a different one
    results = []
    algo = "OnePlusOne"  # for simplifying the test
    rl.agents.TorchAgentFunction._num_test_evaluations = 1  # patch for faster evaluation
    for seed in [random_seed] if skip_seed else [random_seed, random_seed, random_seed + 1]:
        print(f"\nStarting with {seed % 100}")  # useful debug info when this test fails
        xps = list(itertools.islice(maker(seed), 0, 1 if short else 2))
        simplified = [
            Experiment(
                xp.function, algo, budget=2, num_workers=min(2, xp.optimsettings.num_workers), seed=xp.seed
            )
            for xp in xps
        ]
        # compute in any order
        np.random.shuffle(simplified)  # type: ignore
        selector = Selector(data=[xp.run() for xp in simplified])
        results.append(Selector(selector.loc[:, ["loss", "seed", "error"]]))  # elapsed_time can vary...
        assert results[-1].unique("error") == {""}, f"An error was raised during optimization:\n{results[-1]}"
    if skip_seed:
        return
    results[0].assert_equivalent(results[1], f"Non identical outputs for seed={random_seed}")
    np.testing.assert_raises(
        AssertionError,
        results[1].assert_equivalent,
        results[2],
        f"Identical output with different seeds (seed={random_seed})",
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import hashlib
import itertools
import numpy as np
import nevergrad as ng
from nevergrad.common import tools
import nevergrad.common.typing as tp
from .base import ExperimentFunction
from .pbt import PBT as PBT  # pylint: disable=unused-import
from . import utils
from . import corefuncs


class ArtificialVariable:
    # pylint: disable=too-many-instance-attributes,too-many-arguments
    # TODO: refactor, this is not more used for parametrization, so using the
    # Variable framework is not necessary

    def __init__(
        self,
        dimension: int,
        num_blocks: int,
        block_dimension: int,
        translation_factor: float,
        rotation: bool,
        hashing: bool,
        only_index_transform: bool,
        random_state: np.random.RandomState,
        expo: float,
    ) -> None:
        self._dimension = dimension
        self._transforms: tp.List[utils.Transform] = []
        self.rotation = rotation
        self.translation_factor = translation_factor
        self.num_blocks = num_blocks
        self.block_dimension = block_dimension
        self.only_index_transform = only_index_transform
        self.hashing = hashing
        self.dimension = self._dimension
        self.random_state = random_state
        self.expo = expo

    def _initialize(self) -> None:
        """Delayed initialization of the transforms to avoid slowing down the instance creation
        (makes unit testing much faster).
        This functions creates the random transform used upon each block (translation + optional rotation).
        """
        # use random indices for blocks
        indices = self.random_state.choice(
            self._dimension, self.block_dimension * self.num_blocks, replace=False  # type: ignore
        ).tolist()
        indices.sort()  # keep the indices sorted sorted so that blocks do not overlap
        # Caution this is also important for split, so that splitted arrays end un in the same block
        for transform_inds in tools.grouper(indices, n=self.block_dimension):
            self._transforms.append(
                utils.Transform(
                    transform_inds,
                    translation_factor=self.translation_factor,
                    rotation=self.rotation,
                    random_state=self.random_state,
                    expo=self.expo,
                )
            )

    def process(  # pylint: disable=unused-argument
        self, data: tp.ArrayLike, deterministic: bool = True
    ) -> np.ndarray:
        if not self._transforms:
            self._initialize()
        if self.hashing:
            data2 = np.array(data, copy=True)
            for i, y in enumerate(data):
                self.random_state.seed(int(hashlib.md5(str(y).encode()).hexdigest(), 16) % 500000)
                data2[i] = self.random_state.normal(0.0, 1.0)
            data = data2
        data = np.array(data, copy=False)
        output = []
        for transform in self._transforms:
            output.append(data[transform.indices] if self.only_index_transform else transform(data))
        return np.array(output)

    def _short_repr(self) -> str:
        return "Photonics"


class ArtificialFunction(ExperimentFunction):
    """Artificial function object. This allows the creation of functions with different
    dimension and structure to be used for benchmarking in many different settings.

    Parameters
    ----------
    name: str
        name of the underlying function to use (like "sphere" for instance). If a wrong
        name is provided, an error is raised with all existing names.
    block_dimension: int
        the dimension on which the underlying function will be applied.
    num_blocks: int
        the number of blocks of size "block_dimension" on which the underlying function
        will be applied. The number of useful dimension is therefore num_blocks * core_dimension
    useless_variables: int
        the number of additional variables which have no impact on the core function.
        The full dimension of the function is therefore useless_variables + num_blocks * core_dimension
    noise_level: float
        noise level for the additive noise: noise_level * N(0, 1, size=1) * [f(x + N(0, 1, size=dim)) - f(x)]
    noise_dissymmetry: bool
        True if we dissymmetrize the model of noise
    rotation: bool
        whether the block space should be rotated (random rotation)
    hashing: bool
        whether the input data should be hashed. In this case, the function expects an array of size 1 with
        string as element.
    aggregator: str
        how to aggregate the multiple block outputs
    bounded: bool
        bound the search domain to [-5,5]

    Example
    -------
    >>> func = ArtificialFunction("sphere", 5, noise_level=.1)
    >>> x = [1, 2, 1, 0, .5]
    >>> func(x)  # returns a float
    >>> func(x)  # returns a different float since the function is noisy
    >>> func.oracle_call(x)   # returns a float
    >>> func.oracle_call(x)   # returns the same float (no noise for oracles + sphere function is deterministic)
    >>> func2 = ArtificialFunction("sphere", 5, noise_level=.1)
    >>> func2.oracle_call(x)   # returns a different float than before, because a random translation is applied

    Note
    ----
    - The full dimension of the function is available through the dimension attribute.
      Its value is useless_variables + num_blocks * block_dimension
    - The blocks are chosen with random sorted indices (blocks do not overlap)
    - A random translation is always applied to the function at initialization, so that
      instantiating twice the functions will give 2 different functions (unless you use
      seeding)
    - the noise formula is: noise_level * N(0, 1) * (f(x + N(0, 1)) - f(x))
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        name: str,
        block_dimension: int,
        num_blocks: int = 1,
        useless_variables: int = 0,
        noise_level: float = 0,
        noise_dissymmetry: bool = False,
        rotation: bool = False,
        translation_factor: float = 1.0,
        hashing: bool = False,
        aggregator: str = "max",
        split: bool = False,
        bounded: bool = False,
        expo: float = 1.0,
        zero_pen: bool = False,
    ) -> None:
        # pylint: disable=too-many-locals
        self.name = name
        self.expo = expo
        self.translation_factor = translation_factor
        self.zero_pen = zero_pen
        self.constraint_violation: tp.ArrayLike = []
        self._parameters = {x: y for x, y in locals().items() if x not in ["__class__", "self"]}
        # basic checks
        assert noise_level >= 0, "Noise level must be greater or equal to 0"
        if not all(isinstance(x, bool) for x in [noise_dissymmetry, hashing, rotation]):
            raise TypeError("hashing and rotation should be bools")
        for param, mini in [("block_dimension", 1), ("num_blocks", 1), ("useless_variables", 0)]:
            value = self._parameters[param]
            if not isinstance(value, int):
                raise TypeError(f'"{param}" must be an int')
            if value < mini:
                raise ValueError(f'"{param}" must be greater or equal to {mini}')
        if not isinstance(translation_factor, (float, int)):
            raise TypeError(f"Got non-float value {translation_factor}")
        if name not in corefuncs.registry:
            available = ", ".join(self.list_sorted_function_names())
            raise ValueError(f'Unknown core function "{name}". Available names are:\n-----\n{available}')
        # record necessary info and prepare transforms
        self._dimension = block_dimension * num_blocks + useless_variables
        self._func = corefuncs.registry[name]
        # special case
        info = corefuncs.registry.get_info(self._parameters["name"])
        only_index_transform = info.get("no_transform", False)

        array_bounds = dict(upper=5, lower=-5) if bounded else {}
        if not split:
            parametrization: ng.p.Parameter = ng.p.Array(
                shape=(1,) if hashing else (self._dimension,), **array_bounds  # type: ignore
            ).set_name("")
        else:
            assert not hashing
            assert not useless_variables
            arrays = [
                ng.p.Array(shape=(block_dimension,), **array_bounds) for _ in range(num_blocks)  # type: ignore
            ]
            parametrization = ng.p.Instrumentation(*arrays)
            parametrization.set_name("split")
        if noise_level > 0:
            parametrization.function.deterministic = False
        super().__init__(self.noisy_function, parametrization)
        # variable, must come after super().__init__(...) to bind the random_state
        # may consider having its a local random_state instead but less reproducible
        self.transform_var = ArtificialVariable(
            dimension=self._dimension,
            num_blocks=num_blocks,
            block_dimension=block_dimension,
            translation_factor=translation_factor,
            rotation=rotation,
            hashing=hashing,
            only_index_transform=only_index_transform,
            random_state=self._parametrization.random_state,
            expo=self.expo,
        )
        self._aggregator: tp.Callable[[tp.ArrayLike], float] = {  # type: ignore
            "max": np.max,
            "mean": np.mean,
            "sum": np.sum,
        }[aggregator]
        info = corefuncs.registry.get_info(self._parameters["name"])
        # add descriptors
        self.add_descriptors(
            useful_dimensions=block_dimension * num_blocks,
            discrete=any(x in name for x in ["onemax", "leadingones", "jump"]),
        )

    @property
    def dimension(self) -> int:
        # bypass the parametrization one (because of the "hashing" case)  # TODO: remove
        return self._dimension

    @staticmethod
    def list_sorted_function_names() -> tp.List[str]:
        """Returns a sorted list of function names that can be used for the blocks"""
        return sorted(corefuncs.registry)

    def _transform(self, x: tp.ArrayLike) -> np.ndarray:
        data = self.transform_var.process(x)
        return np.array(data)

    def function_from_transform(self, x: np.ndarray) -> float:
        """Implements the call of the function.
        Under the hood, __call__ delegates to oracle_call + add some noise if noise_level > 0.
        """
        results = []
        for block in x:
            results.append(self._func(block))
        try:
            val = float(self._aggregator(results))
            if self.zero_pen:
                val += 1e3 * max(  # type: ignore
                    self.translation_factor / (1e-7 + self.translation_factor + np.linalg.norm(x.flatten()))
                    - 0.75,
                    0.0,
                )
            return val
        except OverflowError:
            return float("inf")

    def evaluation_function(self, *recommendations: ng.p.Parameter) -> float:
        """Implements the call of the function.
        Under the hood, __call__ delegates to oracle_call + add some noise if noise_level > 0.
        """
        assert len(recommendations) == 1, "Should not be a pareto set for a singleobjective function"
        assert not recommendations[0].kwargs
        # we can concatenate since blocks are necessarily sorted
        data = np.concatenate(recommendations[0].args, axis=0)
        data = self._transform(data)
        return self.function_from_transform(data)

    def noisy_function(self, *x: tp.ArrayLike) -> float:
        data = np.concatenate(x, axis=0)  # we can concatenate since blocks are necessarily sorted
        return _noisy_call(
            x=data,
            transf=self._transform,
            func=self.function_from_transform,
            noise_level=self._parameters["noise_level"],
            noise_dissymmetry=self._parameters["noise_dissymmetry"],
            random_state=self._parametrization.random_state,
        )

    def compute_pseudotime(self, input_parameter: tp.ArgsKwargs, loss: tp.Loss) -> float:
        """Delay before returning results in steady state mode benchmarks (fake execution time)"""
        args, kwargs = input_parameter
        assert not kwargs
        if hasattr(self._func, "compute_pseudotime"):
            data = self._transform(np.concatenate(args, axis=0))
            total = 0.0
            for block in data:
                total += self._func.compute_pseudotime(((block,), {}), loss)  # type: ignore
            return total
        return 1.0


def _noisy_call(
    x: np.ndarray,
    transf: tp.Callable[[np.ndarray], np.ndarray],
    func: tp.Callable[[np.ndarray], float],
    noise_level: float,
    noise_dissymmetry: bool,
    random_state: np.random.RandomState,
) -> float:  # pylint: disable=unused-argument
    x_transf = transf(x)
    fx = func(x_transf)
    noise = 0.0
    if noise_level:
        if not noise_dissymmetry or x_transf.ravel()[0] <= 0:
            side_point = transf(x + random_state.normal(0, 1, size=len(x)))
            if noise_dissymmetry:
                noise_level *= 1.0 + x_transf.ravel()[0] * 100.0
            noise = noise_level * random_state.normal(0, 1) * (func(side_point) - fx)
    return fx + noise


class FarOptimumFunction(ExperimentFunction):
    """Very simple 2D norm-1 function with optimal value at (x_optimum, 100)"""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        independent_sigma: bool = True,
        mutable_sigma: bool = True,
        multiobjective: bool = False,
        recombination: str = "crossover",
        optimum: tp.Tuple[int, int] = (80, 100),
    ) -> None:
        assert recombination in ("crossover", "average")
        self._optimum = np.array(optimum, dtype=float)
        parametrization = ng.p.Array(shape=(2,), mutable_sigma=mutable_sigma)
        init = np.array([1.0, 1.0] if independent_sigma else [1.0], dtype=float)
        sigma: tp.Any = ng.p.Array(init=init).set_mutation(exponent=2.0) if mutable_sigma else init
        parametrization.set_mutation(sigma=sigma)
        if recombination == "crossover":
            parametrization = ng.ops.mutations.Crossover()(parametrization)
        self.multiobjective_upper_bounds = np.array(2 * self._optimum) if multiobjective else None
        super().__init__(self._multifunc if multiobjective else self._monofunc, parametrization.set_name(""))  # type: ignore

    def _multifunc(self, x: np.ndarray) -> np.ndarray:
        return np.abs(x - self._optimum)  # type: ignore

    def _monofunc(self, x: np.ndarray) -> float:
        return float(np.sum(self._multifunc(x)))

    def evaluation_function(self, *recommendations: ng.p.Parameter) -> float:
        return min(self._monofunc(x.args[0]) for x in recommendations)

    @classmethod
    def itercases(cls) -> tp.Iterator["FarOptimumFunction"]:
        options = dict(
            independent_sigma=[True, False],
            mutable_sigma=[True, False],
            multiobjective=[True, False],
            recombination=["average", "crossover"],
            optimum=[(0.8, 1), (80, 100), (0.8, 100)],
        )
        keys = sorted(options)
        select = itertools.product(*(options[k] for k in keys))  # type: ignore
        cases = (dict(zip(keys, s)) for s in select)
        return (cls(**c) for c in cases)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Optimization of the FAO crop management model.
Based on 
https://colab.research.google.com/github/thomasdkelly/aquacrop/blob/master/tutorials/AquaCrop_OSPy_Notebook_3.ipynb#scrollTo=YDm931IGNxCb
"""

from nevergrad.parametrization import parameter
from ..base import ExperimentFunction
from ..base import UnsupportedExperiment as UnsupportedExperiment

# pylint: disable=too-many-locals,too-many-statements

# Inspired by
# https://colab.research.google.com/github/thomasdkelly/aquacrop/blob/master/tutorials/AquaCrop_OSPy_Notebook_3.ipynb#scrollTo=YDm931IGNxCb

# In the colab it was:
# from aquacrop.classes import *
# from aquacrop.core import *


class NgAquacrop(ExperimentFunction):
    def __init__(self, num_smts: int, max_irr_seasonal: float) -> None:
        self.num_smts = num_smts
        self.max_irr_seasonal = max_irr_seasonal
        super().__init__(self.loss, parametrization=parameter.Array(shape=(num_smts,)))

    def loss(self, smts):
        try:
            import aquacrop
        except ImportError:
            raise UnsupportedExperiment("Please install aquacrop==0.2 for FAO aquacrop experiments")
        path = aquacrop.core.get_filepath("champion_climate.txt")
        wdf = aquacrop.core.prepare_weather(path)

        def run_model(smts, max_irr_season, year1, year2):
            """
            Function to run model and return results for given set of soil moisture targets.
            """

            maize = aquacrop.classes.CropClass("Maize", PlantingDate="05/01")  # define crop
            loam = aquacrop.classes.SoilClass("ClayLoam")  # define soil
            init_wc = aquacrop.classes.InitWCClass(
                wc_type="Pct", value=[70]
            )  # define initial soil water conditions

            irrmngt = aquacrop.classes.IrrMngtClass(
                IrrMethod=1, SMT=smts, MaxIrrSeason=max_irr_season
            )  # define irrigation management

            # create and run model
            model = aquacrop.core.AquaCropModel(
                f"{year1}/05/01", f"{year2}/10/31", wdf, loam, maize, IrrMngt=irrmngt, InitWC=init_wc
            )
            model.initialize()
            model.step(till_termination=True)
            return model.Outputs.Final

        def evaluate(smts) -> float:  # ,max_irr_season,test=False):
            """
            Function to run model and calculate reward (yield) for given set of soil moisture targets
            """
            max_irr_season = self.max_irr_seasonal
            assert len(smts) == self.num_smts
            out = run_model(smts, max_irr_season, year1=2016, year2=2018)
            # get yields.
            reward = out["Yield (tonne/ha)"].mean()
            return -reward

        return evaluate(smts)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import copy
import scipy.stats
import typing as tp
import numpy as np
import gym

import nevergrad as ng

from nevergrad.parametrization import parameter
from ..base import ExperimentFunction

# pylint: disable=unused-import,import-outside-toplevel


GUARANTEED_GYM_ENV_NAMES = [
    # "ReversedAddition-v0",
    # "ReversedAddition3-v0",
    # "DuplicatedInput-v0",
    # "Reverse-v0",
    "CartPole-v0",
    "CartPole-v1",
    "MountainCar-v0",
    "Acrobot-v1",
    # "Blackjack-v0",
    # "FrozenLake-v0",   # deprecated
    # "FrozenLake8x8-v0",
    "CliffWalking-v0",
    # "NChain-v0",
    # "Roulette-v0",
    "Taxi-v3",
    # "CubeCrash-v0",
    # "CubeCrashSparse-v0",
    # "CubeCrashScreenBecomesBlack-v0",
    # "MemorizeDigits-v0",
]


# We do not use "conformant" which is not consistent with the rest.
CONTROLLERS = [
    "resid_neural",
    "resid_semideep_neural",
    "resid_deep_neural",
    "resid_scrambled_neural",
    "resid_scrambled_semideep_neural",
    "resid_scrambled_deep_neural",
    "resid_noisy_scrambled_neural",
    "resid_noisy_scrambled_semideep_neural",
    "resid_noisy_scrambled_deep_neural",
    "linear",  # Simple linear controller.
    "neural",  # Simple neural controller.
    "deep_neural",  # Deeper neural controller.
    "semideep_neural",  # Deep, but not very deep.
    "structured_neural",  # Structured optimization of a neural net.
    "memory_neural",  # Uses a memory (i.e. recurrent net).
    "deep_memory_neural",
    "stackingmemory_neural",  # Uses a memory and stacks a heuristic and the memory as inputs.
    "deep_stackingmemory_neural",
    "semideep_stackingmemory_neural",
    "extrapolatestackingmemory_neural",  # Same as stackingmemory_neural + suffix-based extrapolation.
    "deep_extrapolatestackingmemory_neural",
    "semideep_extrapolatestackingmemory_neural",
    "semideep_memory_neural",
    "noisy_semideep_neural",
    "noisy_scrambled_semideep_neural",  # Scrambling: why not perturbating the order of variables ?
    "noisy_deep_neural",
    "noisy_scrambled_deep_neural",
    "multi_neural",  # One neural net per time step.
    "noisy_neural",  # Do not start at 0 but at a random point.
    "noisy_scrambled_neural",
    "stochastic_conformant",  # Conformant planning, but still not deterministic.
]


NO_LENGTH = ["ANM", "Blackjack", "CliffWalking", "Cube", "Memorize", "llvm"]


class GymMulti(ExperimentFunction):
    """Class for converting a gym environment, a controller style, and others into a black-box optimization benchmark."""

    @staticmethod
    def get_env_names() -> tp.List[str]:
        # import gym_anm  # noqa

        gym_env_names = []
        max_displays = 10
        for e in gym.envs.registry.all():
            try:
                assert not any(
                    x in str(e.id)
                    for x in "Kelly Copy llvm BulletEnv Minitaur Kuka InvertedPendulumSwingupBulletEnv".split()
                )  # We should have another check than that.
                assert (
                    "RacecarZedBulletEnv-v0" != e.id
                ), "This specific environment causes X11 error when using pybullet_envs."
                assert "CarRacing-v" not in str(e.id), "Pixel based task not supported yet"
                env = gym.make(e.id)
                env.reset()
                env.step(env.action_space.sample())
                a1 = np.asarray(env.action_space.sample())
                a2 = np.asarray(env.action_space.sample())
                a3 = np.asarray(env.action_space.sample())
                a1 = a1 + a2 + a3
                if hasattr(a1, "size"):
                    try:
                        assert a1.size < 15000
                    except Exception:  # pylint: disable=broad-except
                        assert a1.size() < 15000  # type: ignore
                gym_env_names.append(e.id)
            except Exception as exception:  # pylint: disable=broad-except
                max_displays -= 1
                if max_displays > 0:
                    print(f"{e.id} not included in full list because of {exception}.")
                if max_displays == 0:
                    print("(similar issue for other environments)")
        return gym_env_names

    controllers = CONTROLLERS

    ng_gym = [
        # "Reverse-v0",
        "CartPole-v0",
        "CartPole-v1",
        "Acrobot-v1",
        # "FrozenLake-v0",  # deprecated
        # "FrozenLake8x8-v0",
        # "NChain-v0",
        # "Roulette-v0",
    ]

    def wrap_env(self, input_env):
        env = gym.wrappers.TimeLimit(
            env=input_env,
            max_episode_steps=self.num_episode_steps,
        )
        return env

    def create_env(self) -> tp.Any:
        env = gym.make(self.short_name)
        try:
            env.reset()
        except:
            assert False, f"Maybe check if {self.short_name} has a problem in reset / observation."
        return env

    def __init__(
        self,
        name: str = "CartPole-v0",
        control: str = "conformant",
        neural_factor: tp.Optional[int] = 1,
        randomized: bool = True,
        optimization_scale: int = 0,
        greedy_bias: bool = False,
        sparse_limit: tp.Optional[
            int
        ] = None,  # if not None, we penalize solutions with more than sparse_limit weights !=0
    ) -> None:
        self.num_calls = 0
        self.optimization_scale = optimization_scale
        self.stochastic_problem = "stoc" in name
        self.greedy_bias = greedy_bias
        self.sparse_limit = sparse_limit
        if "conformant" in control or control == "linear":
            assert neural_factor is None
        if os.name == "nt":
            raise ng.errors.UnsupportedExperiment("Windows is not supported")
        # self.env = None  # self.create_env() let us have no self.env

        # Build various attributes.
        self.short_name = name  # Just the environment name.
        env = self.create_env()
        self.name = name + "__" + control + "__" + str(neural_factor)
        if sparse_limit is not None:
            self.name += f"__{sparse_limit}"
        if randomized:
            self.name += "_unseeded"
        self.randomized = randomized
        try:
            try:
                self.num_time_steps = env._max_episode_steps  # I know! This is a private variable.
            except AttributeError:  # Second chance! Some environments use self.horizon.
                self.num_time_steps = env.horizon
        except AttributeError:  # Not all environements have a max number of episodes!
            assert any(x in name for x in NO_LENGTH), name
            if "LANM" not in name:  # Most cases: let's say 5000 time steps.
                self.num_time_steps = 200 if control == "conformant" else 5000
            else:  # LANM is a special case with 3000 time steps.
                self.num_time_steps = 3000
        self.gamma = 0.995 if "LANM" in name else 1.0
        self.neural_factor = neural_factor

        # Infer the action space.
        self.arities = []
        if isinstance(env.action_space, gym.spaces.Tuple):
            assert all(
                isinstance(p, gym.spaces.MultiDiscrete) for p in env.action_space
            ), f"{name} has a too complicated structure."
            self.arities = [p.nvec for p in env.action_space]
            if control == "conformant":
                output_dim = sum(len(a) for a in self.arities)
            else:
                output_dim = sum(sum(a) for a in self.arities)
            output_shape = (output_dim,)
            discrete = True
        elif isinstance(env.action_space, gym.spaces.Discrete):
            output_dim = env.action_space.n
            output_shape = (output_dim,)
            discrete = True
            assert output_dim is not None, env.action_space.n
        else:  # Continuous action space
            output_shape = env.action_space.shape
            if output_shape is None:
                output_shape = tuple(np.asarray(env.action_space.sample()).shape)
            # When the shape is not available we might do:
            # output_shape = tuple(np.asarray(env.action_space.sample()).shape)
            discrete = False
            output_dim = np.prod(output_shape)

        self.discrete = discrete

        # Infer the observation space.
        assert env.observation_space is not None or "llvm" in name, "An observation space should be defined."
        if env.observation_space is not None and env.observation_space.dtype == int:
            # Direct inference for corner cases:
            # if "int" in str(type(o)):
            input_dim = env.observation_space.n
            assert input_dim is not None, env.observation_space.n
            self.discrete_input = True
        else:
            input_dim = np.prod(env.observation_space.shape) if env.observation_space is not None else 0
            if input_dim is None:
                o = env.reset()
                input_dim = np.prod(np.asarray(o).shape)
            self.discrete_input = False

        # Infer the action type.
        a = env.action_space.sample()
        self.action_type = type(a)
        self.subaction_type = None
        if hasattr(a, "__iter__"):
            self.subaction_type = type(a[0])

        # Prepare the policy shape.
        if neural_factor is None:
            assert (
                control == "linear" or "conformant" in control
            ), f"{control} has neural_factor {neural_factor}"
            neural_factor = 1
        self.output_shape = output_shape
        self.num_stacking = 1
        self.memory_len = neural_factor * input_dim if "memory" in control else 0
        self.extended_input_len = (input_dim + output_dim) * self.num_stacking if "stacking" in control else 0
        input_dim = input_dim + self.memory_len + self.extended_input_len
        self.extended_input = np.zeros(self.extended_input_len)
        output_dim = output_dim + self.memory_len
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.num_neurons = neural_factor * (input_dim - self.extended_input_len)
        self.num_internal_layers = 1 if "semi" in control else 3
        internal = self.num_internal_layers * (self.num_neurons**2) if "deep" in control else 0
        unstructured_neural_size = (
            output_dim * self.num_neurons + self.num_neurons * (input_dim + 1) + internal,
        )
        neural_size = unstructured_neural_size
        if self.greedy_bias:
            neural_size = (unstructured_neural_size[0] + 1,)
            assert "multi" not in control
            assert "structured" not in control
        assert control in CONTROLLERS or control == "conformant", f"{control} not known as a form of control"
        self.control = control
        if "neural" in control:
            self.first_size = self.num_neurons * (self.input_dim + 1)
            self.second_size = self.num_neurons * self.output_dim
            self.first_layer_shape = (self.input_dim + 1, self.num_neurons)
            self.second_layer_shape = (self.num_neurons, self.output_dim)
        shape_dict = {
            "conformant": (self.num_time_steps,) + output_shape,
            "stochastic_conformant": (self.num_time_steps,) + output_shape,
            "linear": (input_dim + 1, output_dim),
            "multi_neural": (min(self.num_time_steps, 50),) + unstructured_neural_size,
        }
        shape = tuple(map(int, shape_dict.get(control, neural_size)))
        self.policy_shape = shape if "structured" not in control else None

        # Create the parametrization.
        parametrization = parameter.Array(shape=shape).set_name("ng_default")
        if sparse_limit is not None:
            parametrization1 = parameter.Array(shape=shape)
            repetitions = int(np.prod(shape))
            assert isinstance(repetitions, int), f"{repetitions}"
            parametrization2 = ng.p.Choice([0, 1], repetitions=repetitions)  # type: ignore
            parametrization = ng.p.Instrumentation(  # type: ignore
                weights=parametrization1,
                enablers=parametrization2,
            )
            parametrization.set_name("ng_sparse" + str(sparse_limit))
            assert "conformant" not in control and "structured" not in control

        if "structured" in control and "neural" in control and "multi" not in control:
            parametrization = parameter.Instrumentation(  # type: ignore
                parameter.Array(shape=tuple(map(int, self.first_layer_shape))),
                parameter.Array(shape=tuple(map(int, self.second_layer_shape))),
            ).set_name("ng_struct")
        elif "conformant" in control:
            try:
                if env.action_space.low is not None and env.action_space.high is not None:
                    low = np.repeat(np.expand_dims(env.action_space.low, 0), self.num_time_steps, axis=0)
                    high = np.repeat(np.expand_dims(env.action_space.high, 0), self.num_time_steps, axis=0)
                    init = 0.5 * (low + high)
                    parametrization = parameter.Array(init=init)
                    parametrization.set_bounds(low, high)
            except AttributeError:  # Not all env.action_space have a low and a high.
                pass
            if self.subaction_type == int:
                parametrization.set_integer_casting()
            parametrization.set_name("conformant")

        # Now initializing.
        super().__init__(
            self.sparse_gym_multi_function if sparse_limit is not None else self.gym_multi_function,  # type: ignore
            parametrization=parametrization,
        )
        self.greedy_coefficient = 0.0
        self.parametrization.function.deterministic = False
        self.archive: tp.List[tp.Any] = []
        self.mean_loss = 0.0
        self.num_losses = 0

    def evaluation_function(self, *recommendations) -> float:
        """Averages multiple evaluations if necessary."""
        if self.sparse_limit is None:  # Life is simple here, we directly have the weights.
            x = recommendations[0].value
        else:  # Here 0 in the enablers means that the weight is forced to 0.
            # assert np.prod(recommendations[0].value["weights"].shape) == np.prod(recommendations[0].value["enablers"].shape)
            weights = recommendations[0].kwargs["weights"]
            enablers = np.asarray(recommendations[0].kwargs["enablers"])
            assert all(x_ in [0, 1] for x_ in enablers), f"non-binary enablers: {enablers}."
            enablers = enablers.reshape(weights.shape)
            x = weights * enablers
        if not self.randomized:
            return self.gym_multi_function(x, limited_fidelity=False)
        # We want to reduce noise by averaging without
        # spending more than 20% of the whole experiment,
        # hence the line below:
        num = max(self.num_calls // 5, 23)
        # Pb_index >= 0 refers to the test set.
        return np.sum([self.gym_multi_function(x, limited_fidelity=False) for _ in range(num)]) / num

    def softmax(self, a):
        a = np.nan_to_num(a, copy=False, nan=-1e20, posinf=1e20, neginf=-1e20)
        probabilities = np.exp(a - max(a))
        probabilities = probabilities / sum(probabilities)
        return probabilities

    def discretize(self, a, env):
        """Transforms a logit into an int obtained through softmax."""
        if len(self.arities) > 0:
            if self.control == "conformant":
                index = 0
                output = []
                for arities in self.arities:
                    local_output = []
                    for arity in arities:
                        local_output += [min(int(scipy.stats.norm.cdf(a[index]) * arity), arity - 1)]
                        index += 1
                    output += [local_output]
                assert index == len(a)
                return np.array(output)
            else:
                index = 0
                output = []
                for arities in self.arities:
                    local_output = []
                    for arity in arities:
                        local_output += [
                            int(
                                list(np.random.multinomial(1, self.softmax(a[index : index + arity]))).index(
                                    1
                                )
                            )
                        ]
                        index += arity
                    output += [local_output]
                assert index == len(a)
                return np.array(output)

        if self.greedy_bias:
            a = np.asarray(a, dtype=np.float32)
            for i, action in enumerate(range(len(a))):
                tmp_env = copy.deepcopy(env)
                _, r, _, _ = tmp_env.step(action)
                a[i] += self.greedy_coefficient * r
        probabilities = self.softmax(a)
        assert sum(probabilities) <= 1.0 + 1e-7, f"{probabilities} with greediness {self.greedy_coefficient}."
        return int(list(np.random.multinomial(1, probabilities)).index(1))

    def neural(self, x: np.ndarray, o: np.ndarray):
        """Applies a neural net parametrized by x to an observation o. Returns an action or logits of actions."""
        if self.greedy_bias:
            assert "multi" not in self.control
            assert "structured" not in self.control
            self.greedy_coefficient = x[-1:]  # We have decided that we can not have two runs in parallel.
            x = x[:-1]
        o = o.ravel()
        my_scale = 2**self.optimization_scale
        if "structured" not in self.name and self.optimization_scale != 0:
            x = np.asarray(my_scale * x, dtype=np.float32)
        if self.control == "linear":
            # The linear case is simplle.
            output = np.matmul(o, x[1:, :])
            output += x[0]
            return output.reshape(self.output_shape), np.zeros(0)
        if "structured" not in self.control:
            # If not structured then we split into two matrices.
            first_matrix = x[: self.first_size].reshape(self.first_layer_shape) / np.sqrt(len(o))
            second_matrix = x[self.first_size : (self.first_size + self.second_size)].reshape(
                self.second_layer_shape
            ) / np.sqrt(self.num_neurons)
        else:
            # In the structured case we should have two entries with the right shapes.
            assert len(x) == 2
            first_matrix = np.asarray(x[0][0])
            second_matrix = np.asarray(x[0][1])
            assert (
                first_matrix.shape == self.first_layer_shape
            ), f"{first_matrix} does not match {self.first_layer_shape}"
            assert (
                second_matrix.shape == self.second_layer_shape
            ), f"{second_matrix} does not match {self.second_layer_shape}"
        if "resid" in self.control:
            first_matrix += my_scale * np.eye(*first_matrix.shape)
            second_matrix += my_scale * np.eye(*second_matrix.shape)
        assert len(o) == len(first_matrix[1:]), f"{o.shape} coming in matrix of shape {first_matrix.shape}"
        output = np.matmul(o, first_matrix[1:])
        if "deep" in self.control:
            # The deep case must be split into several layers.
            current_index = self.first_size + self.second_size
            internal_layer_size = self.num_neurons**2
            s = (self.num_neurons, self.num_neurons)
            for _ in range(self.num_internal_layers):
                output = np.tanh(output)
                layer = x[current_index : current_index + internal_layer_size].reshape(s)
                if "resid" in self.control:
                    layer += my_scale * np.eye(*layer.shape)
                output = np.matmul(output, layer) / np.sqrt(self.num_neurons)
                current_index += internal_layer_size
            assert current_index == len(x)
        output = np.matmul(np.tanh(output + first_matrix[0]), second_matrix)
        return output[self.memory_len :].reshape(self.output_shape), output[: self.memory_len]

    def sparse_gym_multi_function(
        self,
        weights: np.ndarray,
        enablers: np.ndarray,
        limited_fidelity: bool = False,
    ) -> float:
        assert all(x_ in [0, 1] for x_ in enablers)
        x = weights * enablers
        loss = self.gym_multi_function(
            x,
            limited_fidelity=limited_fidelity,
        )
        sparse_penalty = 0
        if self.sparse_limit is not None:  # Then we penalize the weights above the threshold "sparse_limit".
            sparse_penalty = (1 + np.abs(loss)) * max(0, np.sum(enablers) - self.sparse_limit)
        return loss + sparse_penalty

    def gym_multi_function(self, x: np.ndarray, limited_fidelity: bool = False) -> float:
        """Do a simulation with parametrization x and return the result.

        Parameters:
            limited_fidelity: bool
                whether we use a limited version for the beginning of the training.
        """
        self.num_calls += 1
        # The deterministic case consists in considering the average of 7 fixed seeds.
        # The conformant case is using 1 randomized seed (unlesss we requested !randomized).
        num_simulations = 7 if self.control != "conformant" and not self.randomized else 1
        loss = 0
        for simulation_index in range(num_simulations):
            loss += self.gym_simulate(
                x,
                seed=simulation_index
                if not self.randomized
                else self.parametrization.random_state.randint(500000),
                limited_fidelity=limited_fidelity,
            )
        return loss / num_simulations

    def action_cast(self, a, env):
        """Transforms an action into an action of type as expected by the gym step function."""
        if self.action_type == tuple:
            a = self.discretize(a, env)
            return tuple(a)
        if type(a) == np.float64:
            a = np.asarray((a,))
        if self.discrete:
            a = self.discretize(a, env)
        else:
            if type(a) != self.action_type:  # , f"{a} does not have type {self.action_type}"
                a = self.action_type(a)
            try:
                if env.action_space.low is not None and env.action_space.high is not None:
                    # Projection to [0, 1]
                    a = 0.5 * (1.0 + np.tanh(a))
                    # Projection to the right space.
                    a = env.action_space.low + (env.action_space.high - env.action_space.low) * a
            except AttributeError:
                pass  # Sometimes an action space has no low and no high.
            if self.subaction_type is not None:
                if type(a) == tuple:
                    a = tuple(int(_a + 0.5) for _a in a)
                else:
                    for i in range(len(a)):
                        a[i] = self.subaction_type(a[i])
        if not np.isscalar(a):
            a = np.asarray(a, dtype=env.action_space.sample().dtype)
        assert type(a) == self.action_type, f"{a} should have type {self.action_type} "
        # assert env.action_space.contains(env.action_space.sample())
        try:
            assert env.action_space.contains(a), (
                f"In {self.name}, high={env.action_space.high} low={env.action_space.low} {a} "
                f"is not sufficiently close to {[env.action_space.sample() for _ in range(10)]}"
                f"Action space = {env.action_space} (sample has type {type(env.action_space.sample())})"
                f"and a={a} with type {type(a)}"
            )
        except AttributeError:
            pass  # Not all env can do "contains".
        return a

    def step(self, a, env):
        """Apply an action.

        We have a step on top of Gym's step for possibly storing some statistics."""
        o, r, done, info = env.step(
            a
        )  # We work on env... we can not have two threads working on the same function.
        return o, r, done, info

    def heuristic(self, o, current_observations):
        """Returns a heuristic action in the given context.

        Parameters:
           o: gym observation
               new observation
           current_observations: list of gym observations
               current observations up to the present time step.
        """
        current_observations = np.asarray(current_observations + [o], dtype=np.float32)
        self.archive = [
            self.archive[i] for i in range(len(self.archive)) if self.archive[i][2] <= self.mean_loss
        ]
        self.archive = sorted(self.archive, key=lambda trace: -len(trace[0]))
        for trace in self.archive:
            to, ta, _ = trace
            assert len(to) == len(ta)
            if len(current_observations) > len(to) and "extrapolate" not in self.control:
                continue
            to = np.asarray(to[(-len(current_observations)) :], dtype=np.float32)
            # if all((_to - _o) for _to, _o in zip(to, current_observations)) <= 1e-7:
            if np.array_equal(to, current_observations):
                return np.asarray(ta[len(current_observations) - 1], dtype=np.float32)
        return None

    def gym_simulate(
        self,
        x: np.ndarray,
        seed: int,
        limited_fidelity: bool = True,
    ):
        """Single simulation with parametrization x."""
        current_time_index = 0
        current_reward = 0.0
        current_observations: tp.List[tp.Any] = []
        current_actions: tp.List[tp.Any] = []
        try:
            if self.policy_shape is not None:
                x = x.reshape(self.policy_shape)
        except:
            assert False, f"x has shape {x.shape} and needs {self.policy_shape} for control {self.control}"
        assert seed == 0 or self.control != "conformant" or self.randomized
        env = self.create_env()
        env.seed(seed=seed)
        o = env.reset()
        control = self.control
        if (
            "conformant" in control
        ):  # Conformant planning: we just optimize a sequence of actions. No reactivity.
            return self.gym_conformant(x, env)
        if "scrambled" in control:  # We shuffle the variables, typically so that progressive methods optimize
            # everywhere in parallel instead of focusing on one single layer for years.
            x = x.copy()
            np.random.RandomState(1234).shuffle(x)
        if "noisy" in control:  # We add a randomly chosen but fixed perturbation of the x, i.e. we do not
            # start at 0.
            x = x + 0.01 * np.random.RandomState(1234).normal(size=x.shape)
        reward = 0.0
        memory = np.zeros(self.memory_len)
        for i in range(self.num_time_steps):
            # Actual loop over time steps!
            if self.discrete_input:
                obs = np.zeros(shape=self.input_dim - self.extended_input_len - len(memory))
                obs[o] = 1
                o = obs
            previous_o = np.asarray(o)
            o = np.concatenate([previous_o.ravel(), memory.ravel(), self.extended_input])
            assert len(o) == self.input_dim, (
                f"o has shape {o.shape} whereas input_dim={self.input_dim} "
                f"({control} / {env} {self.name})"
            )
            a, memory = self.neural(x[i % len(x)] if "multi" in control else x, o)
            a = self.action_cast(a, env)
            try:
                o, r, done, _ = self.step(a, env)  # Outputs = observation, reward, done, info.
                current_time_index += 1
                if "multifidLANM" in self.name and current_time_index > 500 and limited_fidelity:
                    done = True
                current_reward *= self.gamma
                current_reward += r
                current_observations += [np.asarray(o).copy()]
                current_actions += [np.asarray(a).copy()]
                if (
                    done and "stacking" in self.control
                ):  # Only the method which do a stacking of heuristic + memory into the
                    # observation need archiving.
                    self.archive_observations(current_actions, current_observations, current_reward)
            except AssertionError:  # Illegal action.
                return 1e20 / (1.0 + i)  # We encourage late failures rather than early failures.
            if "stacking" in control:
                attention_a = self.heuristic(
                    o, current_observations
                )  # Best so far, or something like that heuristically derived.
                a = attention_a if attention_a is not None else 0.0 * np.asarray(a)
                previous_o = previous_o.ravel()
                additional_input = np.concatenate([np.asarray(a).ravel(), previous_o])
                shift = len(additional_input)
                self.extended_input[: (len(self.extended_input) - shift)] = self.extended_input[shift:]
                self.extended_input[(len(self.extended_input) - shift) :] = additional_input
            reward += r
            if done:
                break
        return -reward

    def gym_conformant(self, x: np.ndarray, env: tp.Any):
        """Conformant: we directly optimize inputs, not parameters of a policy."""
        reward = 0.0
        for _, a in enumerate(x):
            a = self.action_cast(a, env)
            # try:
            _, r, done, _ = self.step(a, env)  # Outputs = observation, reward, done, info.
            # except AssertionError:  # Illegal action.
            #    return 1e20 / (1.0 + i)  # We encourage late failures rather than early failures.
            reward *= self.gamma
            reward += r
            if done:
                break
        # env.render()  if you want to display.
        return -reward

    def archive_observations(self, current_actions, current_observations, current_reward):
        self.num_losses += 1
        tau = 1.0 / self.num_losses
        self.mean_loss = (
            ((1.0 - tau) * self.mean_loss + tau * current_reward)
            if self.mean_loss is not None
            else current_reward
        )
        found = False
        for trace in self.archive:
            to, _, _ = trace
            if np.array_equal(
                np.asarray(current_observations, dtype=np.float32),
                np.asarray(to, dtype=np.float32),
            ):
                found = True
                break
        if not found:
            # Risky: this code assumes that the object is used only in a single run.
            self.archive += [(current_observations, current_actions, current_reward)]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# Ideas and reference implementations from:
# - Pascal Kerschke, University of Muenster
# - Marcus Gallagher, University of Queensland
# - Mike Preuss, LIACS, Leiden University

import numpy as np
import pandas as pd
import scipy.spatial
from nevergrad.parametrization import parameter as p
import nevergrad.common.typing as tp
from ..base import ExperimentFunction
from . import datasets


def _kmeans_distance(points: np.ndarray, centers: np.ndarray) -> float:
    """Computes the distance between points and centers
    after affecting each points to the closest center.
    """
    assert points.shape[1] == centers.shape[1]
    distances = np.sum((points[:, :, None] - centers.T[None, :, :]) ** 2, axis=1)
    affect = np.min(distances, axis=1)
    return float(np.sum(affect))


class Clustering(ExperimentFunction):
    """Cost function of a clustering problem.

    Parameters
    ----------
    points: np.ndarray
        k x n array where k is the number of points and n their coordinates
    num_clusters: int
        number of clusters to find
    """

    def __init__(self, points: np.ndarray, num_clusters: int, rescale: bool = True) -> None:
        self.num_clusters = num_clusters
        self._points = np.array(points, copy=True)
        if rescale:
            self._points -= np.mean(self._points, axis=0, keepdims=True)
            self._points /= np.std(self._points, axis=0, keepdims=True)
        super().__init__(self._compute_distance, p.Array(shape=(num_clusters, points.shape[1])))

    @classmethod
    def from_mlda(cls, name: str, num_clusters: int, rescale: bool = True) -> "Clustering":
        """Clustering problem defined in
        Machine Learning and Data Analysis (MLDA) Problem Set v1.0, Gallagher et al., 2018, PPSN
        https://drive.google.com/file/d/1fc1sVwoLJ0LsQ5fzi4jo3rDJHQ6VGQ1h/view

        Parameters
        ----------
        name: str
            name of the dataset ("Ruspini", or "German towns")
        num_clusters: int
            number of clusters to find

        Note
        ----
        The MLDA problems are P1.a Clustering("Ruspini", 5) and  P1.b Clustering("German towns", 10)
        """
        assert name in ["Ruspini", "German towns"]
        points = datasets.get_data(name)
        pb = cls(points=points, num_clusters=num_clusters, rescale=rescale)
        pb.add_descriptors(name=name)
        return pb

    def _compute_distance(self, centers: np.ndarray) -> float:
        """Sum of minimum squared distances to closest centroid
        centers must be of size num_clusters x n
        """
        return _kmeans_distance(self._points, centers)


class Perceptron(ExperimentFunction):
    """Perceptron function

    Parameters
    ----------
    x: np.ndarray
        the input data
    y: np.ndarray
        the data to predict from the input data
    """

    def __init__(self, x: np.ndarray, y: np.ndarray) -> None:
        assert x.ndim == 1
        assert y.ndim == 1
        self._x = x
        self._y = y
        super().__init__(self._compute_loss, p.Array(shape=(10,)))

    @classmethod
    def from_mlda(cls, name: str) -> "Perceptron":
        """Perceptron problem defined in
        Machine Learning and Data Analysis (MLDA) Problem Set v1.0, Gallagher et al., 2018, PPSN
        https://drive.google.com/file/d/1fc1sVwoLJ0LsQ5fzi4jo3rDJHQ6VGQ1h/view

        Parameters
        ----------
        name: str
            name of the dataset (among "quadratic", "sine", "abs" and "heaviside")

        Note
        ----
        quadratic is coined P2.a, sine P2.b, abs P2.c and heaviside P2.d
        """
        data = datasets.make_perceptron_data(name)
        pb = cls(data[:, 0], data[:, 1])
        pb.add_descriptors(name=name)
        return pb

    def apply(self, parameters: tp.ArrayLike) -> np.ndarray:
        """Apply the perceptron transform to x using the provided parameters

        Parameters
        ----------
        parameters: ArrayLike
            parameters of the perceptron

        Returns
        -------
        np.ndarray
            transformed data
        """
        parameters = np.array(parameters, copy=False)
        assert parameters.shape == (10,)
        tmp = np.tanh(self._x[:, None] * parameters[None, :3] + parameters[None, 3:6])
        tmp *= parameters[None, 6:9]
        output: np.ndarray = np.sum(tmp, axis=1) + parameters[-1]
        return output

    def _compute_loss(self, x: tp.ArrayLike) -> float:
        """Compute perceptron"""
        gx = self.apply(x)
        return float(np.mean((gx - self._y) ** 2))


class SammonMapping(ExperimentFunction):
    """Sammon mapping function"""

    def __init__(self, proximity_array: np.ndarray) -> None:
        self._proximity = proximity_array
        self._proximity_2 = self._proximity**2
        self._proximity_2[
            self._proximity_2 == 0
        ] = 1  # avoid ZeroDivision (for diagonal terms, or identical points)
        super().__init__(self._compute_distance, p.Array(shape=(self._proximity.shape[0], 2)))

    @classmethod
    def from_mlda(cls, name: str, rescale: bool = False) -> "SammonMapping":
        """Mapping problem defined in
        Machine Learning and Data Analysis (MLDA) Problem Set v1.0, Gallagher et al., 2018, PPSN
        https://drive.google.com/file/d/1fc1sVwoLJ0LsQ5fzi4jo3rDJHQ6VGQ1h/view

        Parameters
        ----------
        name: str
            name of the dataset (initially among "Virus", "Employees", but Employees dataset is not
            available online anymore)

        Notes
        -----
        - "Virus" dataset is P3.a and "Employees" dataset is P3.b
        - for "Employees", we use the online proximity matrix
        - for "Virus", we compute a proximity matrix from raw data (no normalization)
        """
        assert name in ["Virus", "Employees"], f"Unkwnown name {name}"
        assert name != "Employees", "The Employees dataset is not available anymore"
        raw_data: pd.DataFrame = datasets.get_data(name)  # type: ignore
        if name == "Employees":
            if rescale:
                raise ValueError("Impossible to rescale with 'Employees'")
            proximity = np.array(raw_data.iloc[:, 1:], dtype=float)  # this is alreary the proximity matrix
            # in this proximity matrix, (5, 31) and (8, 41) are identical
        else:
            if rescale:
                raw_data -= np.mean(raw_data, axis=0, keepdims=True)
                raw_data /= np.std(raw_data, axis=0, keepdims=True)
            proximity = scipy.spatial.distance_matrix(
                raw_data, raw_data
            )  # for Virus, the proximity matrix must be computed
        pb = cls(proximity)
        pb.add_descriptors(name=name, rescale=rescale)
        return pb

    @classmethod
    def from_2d_circle(cls, num_points: int = 12) -> "SammonMapping":
        """Simple test case where the points are in a 2d circle."""
        idata = np.exp(np.linspace(0, 2 * np.pi, num_points) * 1j)
        data = np.zeros((num_points, 2))
        data[:, 0] = np.real(idata)
        data[:, 1] = np.imag(idata)
        instance = cls(scipy.spatial.distance_matrix(data, data))
        instance.add_descriptors(name="circle", num_points=num_points)
        return instance

    def _compute_distance(self, x: np.ndarray) -> float:
        """Compute the Sammon mapping metric for the input data"""
        proximity = scipy.spatial.distance_matrix(x, x)
        norm_prox = (proximity - self._proximity) ** 2 / self._proximity_2
        return float(np.sum(norm_prox.ravel()))


class Landscape(ExperimentFunction):
    """Planet Earth Landscape problem defined in
    Machine Learning and Data Analysis (MLDA) Problem Set v1.0, Gallagher et al., 2018, PPSN
    https://drive.google.com/file/d/1fc1sVwoLJ0LsQ5fzi4jo3rDJHQ6VGQ1h/view
    and reverted to look for the minimum value (0). This is problem P4.

    Parameters
    ----------
    transform: None, "gaussian" or "square"
        whether use the image [0, 4319]x[0, 2159] (None) or to use a Gaussian transform.

    Note
    ----
    - the initial image is 4320x2160
    - sampling outside yields a +inf value (except for Gaussian, since large values are mapped to border indices)
    - the image is actually a variant of the one proposed in the article. Indeed, this image
      has a much better z-resolution. It is not directly proportional to the altitude though
      since it is an artificial rescaling to greyscale of a color image.
    """

    def __init__(self, transform: tp.Optional[str] = None) -> None:
        super().__init__(
            self._get_pixel_value, p.Instrumentation(p.Scalar(), p.Scalar()).set_name("standard")
        )
        self._image = datasets.get_data("Landscape")
        if transform == "gaussian":
            variables = list(p.TransitionChoice(list(range(x))) for x in self._image.shape)
            self.parametrization = p.Instrumentation(*variables).set_name("gaussian")
        elif transform == "square":
            stds = (np.array(self._image.shape) - 1.0) / 2.0
            variables2 = list(p.Scalar(init=s).set_mutation(sigma=s) for s in stds)
            self.parametrization = p.Instrumentation(*variables2).set_name(
                "square"
            )  # maybe buggy, try again?
        elif transform is not None:
            raise ValueError(f"Unknown transform {transform}")
        self._descriptors.pop("transform", None)  # automatically added by __new__, but not needed
        self._max = float(self._image.max())

    def _get_pixel_value(self, x: float, y: float) -> float:
        z = np.round([x, y])
        if np.min(z) < 0 or z[0] >= self._image.shape[0] or z[1] >= self._image.shape[1]:
            return float("inf")  # could propose other limit conditions
        return float(self._max - self._image[int(z[0]), int(z[1])])



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import typing as tp
import numpy as np
from nevergrad.common import testing
from . import game


@testing.parametrized(**{name: (name,) for name in game._Game().get_list_of_games()})
def test_games(name: str) -> None:
    dimension = game._Game().play_game(name)
    res: tp.List[tp.Any] = []
    for k in range(200):
        res.append(game._Game().play_game(name, np.random.uniform(0, 1, dimension), None))
        score = float(sum(1 if r == 2 else 0 if r == 1 else 0.5 for r in res)) / len(res)
        if k >= 20 and 0.2 <= score <= 0.8:
            break
    assert score >= 0.1
    assert score <= 0.9
    assert any(res), "All ties"
    function = game.Game(name)
    function(function.parametrization.random_state.normal(size=function.dimension))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pytest
import numpy as np
import nevergrad as ng
from nevergrad.common.tools import flatten
from .optimizerlib import registry
from .externalbo import _hp_parametrization_to_dict, _hp_dict_to_parametrization


@pytest.mark.parametrize(  # type: ignore
    "parametrization,has_transform",
    [
        (ng.p.Choice(list(range(10))), True),
        (ng.p.Scalar(lower=0, upper=1), True),
        (ng.p.Scalar(lower=0, upper=10).set_integer_casting(), True),
        (ng.p.Log(lower=1e-3, upper=1e3), True),
        (ng.p.Array(init=np.zeros(10)), True),
        (ng.p.Instrumentation(ng.p.Scalar(lower=0, upper=1), a=ng.p.Choice(list(range(10)))), False),
        (
            ng.p.Instrumentation(
                a=ng.p.Choice([ng.p.Scalar(lower=0, upper=1), ng.p.Scalar(lower=100, upper=1000)])
            ),
            True,
        ),
        (
            ng.p.Instrumentation(
                a=ng.p.Choice(
                    [
                        ng.p.Choice(list(range(10))),
                        ng.p.Scalar(lower=0, upper=1),
                    ]
                )
            ),
            False,
        ),
        (
            ng.p.Instrumentation(
                a=ng.p.Choice(
                    [
                        ng.p.Instrumentation(
                            b=ng.p.Choice(list(range(10))), c=ng.p.Log(lower=1e-3, upper=1e3)
                        ),
                        ng.p.Instrumentation(
                            d=ng.p.Scalar(lower=0, upper=1), e=ng.p.Log(lower=1e-3, upper=1e3)
                        ),
                    ]
                )
            ),
            False,
        ),
    ],
)
def test_hyperopt(parametrization, has_transform) -> None:
    optim1 = registry["HyperOpt"](parametrization=parametrization, budget=5)
    optim2 = registry["HyperOpt"](parametrization=parametrization.copy(), budget=5)
    for it in range(4):
        cand = optim1.ask()
        optim1.tell(cand, 0)  # Tell asked
        del cand._meta["trial_id"]
        optim2.tell(cand, 0)  # Tell not asked
        assert flatten(optim1.trials._dynamic_trials[it]["misc"]["vals"]) == pytest.approx(  # type: ignore
            flatten(optim2.trials._dynamic_trials[it]["misc"]["vals"])  # type: ignore
        )

    assert optim1.trials.new_trial_ids(1) == optim2.trials.new_trial_ids(1)  # type: ignore
    assert optim1.trials.new_trial_ids(1)[0] == (it + 2)  # type: ignore
    assert (optim1._transform is not None) == has_transform  # type: ignore

    # Test parallelization
    opt = registry["HyperOpt"](parametrization=parametrization, budget=30, num_workers=5)
    for k in range(40):
        cand = opt.ask()
        if not k:
            opt.tell(cand, 1)


@pytest.mark.parametrize(  # type: ignore
    "parametrization,values",
    [
        (
            ng.p.Instrumentation(
                a=ng.p.Choice([ng.p.Choice(list(range(10))), ng.p.Scalar(lower=0, upper=1)])
            ),
            [
                (((), {"a": 0.5}), {"a": [1], "a__1": [0.5]}, {"args": {}, "kwargs": {"a": 0.5}}),
                (((), {"a": 1}), {"a": [0], "a__0": [1]}, {"args": {}, "kwargs": {"a": 1}}),
            ],
        ),
        (
            ng.p.Instrumentation(ng.p.Scalar(lower=0, upper=1), a=ng.p.Choice(list(range(10)))),
            [
                (((0.5,), {"a": 3}), {"0": [0.5], "a": [3]}, {"args": {"0": 0.5}, "kwargs": {"a": 3}}),
                (((0.99,), {"a": 0}), {"0": [0.99], "a": [0]}, {"args": {"0": 0.99}, "kwargs": {"a": 0}}),
            ],
        ),
        (
            ng.p.Instrumentation(
                a=ng.p.Choice(
                    [
                        ng.p.Instrumentation(
                            b=ng.p.Choice(list(range(10))), c=ng.p.Log(lower=1e-3, upper=1e3)
                        ),
                        ng.p.Instrumentation(
                            d=ng.p.Scalar(lower=0, upper=1), e=ng.p.Log(lower=1e-3, upper=1e3)
                        ),
                    ]
                )
            ),
            [
                (
                    ((), {"a": ((), {"d": 0.5, "e": 1.0})}),
                    {"a": [1], "d": [0.5], "e": [1.0]},
                    {"args": {}, "kwargs": {"a": {"args": {}, "kwargs": {"d": 0.5, "e": 1.0}}}},
                ),
                (
                    ((), {"a": ((), {"b": 0, "c": 0.014})}),
                    {"a": [0], "b": [0], "c": [0.014]},
                    {"args": {}, "kwargs": {"a": {"args": {}, "kwargs": {"b": 0, "c": 0.014}}}},
                ),
            ],
        ),
    ],
)
def test_hyperopt_helpers(parametrization, values):
    for val, dict_val, hyperopt_val in values:
        parametrization.value = val
        assert flatten(_hp_parametrization_to_dict(parametrization)) == pytest.approx(flatten(dict_val))
        assert flatten(_hp_dict_to_parametrization(hyperopt_val)) == pytest.approx(
            flatten(parametrization.value)
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# (C) Copyright 2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This work is implemented under the Formulations and Computational Engineering (FORCE) project within Horizon 2020
# (`NMBP-23-2016/721027 <https://www.the-force-project.eu>`_).
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np

from .hypervolume import (
    VectorNode,
    VectorLinkedList,
    HypervolumeIndicator,
)


def test_initialize_empty_node() -> None:
    dim = 4
    node = VectorNode(dim)

    assert isinstance(node.coordinates, np.ndarray)
    for entry in node.next:
        assert entry is node
    for entry in node.prev:
        assert entry is node

    assert list(node.area) == [0.0] * dim
    assert list(node.volume) == [0.0] * dim
    assert str(node) == "None"


def test_initialize_node() -> None:
    dim = 4
    coordinates = [1.0, 2.0, 3.0]
    node = VectorNode(dim, coordinates=coordinates)

    assert isinstance(node.coordinates, np.ndarray)
    assert list(node.coordinates) == coordinates
    for entry in node.next:
        assert entry is node
    for entry in node.prev:
        assert entry is node

    assert list(node.area) == [0.0] * dim
    assert list(node.volume) == [0.0] * dim
    assert str(node) == "[1. 2. 3.]"

    node.configure_area(0)
    assert node.area[0] == 1.0
    assert node.area[1] == 0.0
    assert node.area[2] == 0.0

    node.configure_area(1)
    assert node.area[0] == 1.0
    assert node.area[1] == -1.0
    assert node.area[2] == 0.0

    node.configure_area(2)
    assert node.area[0] == 1.0
    assert node.area[1] == -1.0
    assert node.area[2] == 2.0


def test_initialize_linked_list() -> None:
    dim = 4
    multilist = VectorLinkedList(dimension=dim)

    assert dim == multilist.dimension
    assert isinstance(multilist.sentinel, VectorNode)
    assert len(multilist.sentinel.prev) == 4
    assert len(multilist.sentinel.next) == 4
    assert len(multilist) == 4

    for d in range(dim):
        assert multilist.sentinel is multilist.sentinel.next[d]
        assert multilist.sentinel is multilist.sentinel.prev[d]

    assert len(multilist.sentinel.next) == len(multilist.sentinel.prev)
    assert len(multilist.sentinel.next) == len(multilist.sentinel.next[0].next)

    assert str(multilist) == "\n".join([str([])] * dim)


def test_append() -> None:
    dim = 4
    multilist = VectorLinkedList(dimension=dim)

    new_node = VectorNode(dim)
    multilist.append(new_node, 0)

    for i in range(1, dim):
        assert new_node.next[i] is new_node
        assert new_node.prev[i] is new_node
        assert multilist.sentinel.next[i] is multilist.sentinel
        assert multilist.sentinel.prev[i] is multilist.sentinel

    assert new_node.next[0] is multilist.sentinel
    assert new_node.prev[0] is multilist.sentinel
    assert multilist.sentinel.next[0] is new_node
    assert multilist.sentinel.prev[0] is new_node

    another_node = VectorNode(dim)
    multilist.append(another_node, 0)
    for i in range(1, dim):
        assert new_node.next[i] is new_node
        assert new_node.prev[i] is new_node
        assert multilist.sentinel.next[i] is multilist.sentinel
        assert multilist.sentinel.prev[i] is multilist.sentinel

    assert new_node.next[0] is another_node
    assert new_node.prev[0] is multilist.sentinel
    assert multilist.sentinel.next[0] is new_node
    assert multilist.sentinel.prev[0] is another_node


def test_extend() -> None:
    dim = 1
    multilist = VectorLinkedList(dimension=dim)
    another_multilist = VectorLinkedList(dimension=dim)

    new_node = VectorNode(dim)
    another_node = VectorNode(dim)

    multilist.append(new_node, 0)
    multilist.append(another_node, 0)

    another_multilist.extend([new_node, another_node], 0)
    assert another_multilist.chain_length(0) == 2
    assert another_multilist.sentinel.next[0] is multilist.sentinel.next[0]
    assert another_multilist.sentinel.next[0].next[0] is multilist.sentinel.next[0].next[0]


def test_chain_length() -> None:
    dim = 3
    multilist = VectorLinkedList(dimension=dim)

    new_node = VectorNode(dim)
    multilist.append(new_node, 0)
    assert multilist.chain_length(0) == 1
    assert multilist.chain_length(1) == 0
    assert multilist.chain_length(2) == 0

    another_node = VectorNode(dim)
    multilist.append(another_node, 0)
    assert multilist.chain_length(0) == 2
    assert multilist.chain_length(1) == 0
    assert multilist.chain_length(2) == 0

    multilist.append(another_node, 1)
    assert multilist.chain_length(0) == 2
    assert multilist.chain_length(1) == 1
    assert multilist.chain_length(2) == 0

    multilist.append(new_node, 2)
    assert multilist.chain_length(0) == 2
    assert multilist.chain_length(1) == 1
    assert multilist.chain_length(2) == 1


def test_pop() -> None:
    dim = 4
    multilist = VectorLinkedList(dimension=dim)

    new_node = VectorNode(dim)
    multilist.append(new_node, 0)

    popped_node = multilist.pop(new_node, 0 + 1)
    assert popped_node is new_node
    assert new_node.next[0] is multilist.sentinel
    assert new_node.prev[0] is multilist.sentinel
    for i in range(dim):
        assert multilist.sentinel.next[i] is multilist.sentinel
        assert multilist.sentinel.prev[i] is multilist.sentinel


def test_reinsert() -> None:
    dim = 2
    multilist = VectorLinkedList(dimension=dim)

    new_node = VectorNode(dim)
    another_node = VectorNode(dim)

    multilist.append(new_node, 0)
    multilist.append(another_node, 0)

    multilist.append(another_node, 1)
    multilist.append(new_node, 1)

    popped_node = multilist.pop(new_node, 1 + 1)

    multilist.reinsert(new_node, 0 + 1)
    assert multilist.chain_length(0) == 2
    assert multilist.chain_length(1) == 1
    assert new_node.next[0] is another_node
    assert new_node.prev[0] is multilist.sentinel
    assert another_node.prev[0] is new_node
    assert another_node.next[0] is multilist.sentinel
    assert another_node.prev[1] is multilist.sentinel
    assert another_node.next[1] is multilist.sentinel

    multilist.reinsert(popped_node, 1 + 1)
    assert multilist.chain_length(0) == 2
    assert multilist.chain_length(1) == 2
    assert another_node.prev[1] is multilist.sentinel
    assert another_node.next[1] is new_node
    assert new_node.prev[1] is another_node
    assert new_node.next[1] is multilist.sentinel


def test_iterate() -> None:
    dim = 1
    multilist = VectorLinkedList(dimension=dim)

    new_node = VectorNode(dim)
    another_node = VectorNode(dim)

    multilist.append(new_node, 0)
    multilist.append(another_node, 0)
    gen = multilist.iterate(0)
    assert next(gen) is new_node
    assert next(gen) is another_node

    yet_another_node = VectorNode(dim)
    multilist.append(yet_another_node, 0)
    gen = multilist.iterate(0, start=another_node)
    assert next(gen) is another_node
    assert next(gen) is yet_another_node


def test_reverse_iterate() -> None:
    dim = 1
    multilist = VectorLinkedList(dimension=dim)

    new_node = VectorNode(dim)
    another_node = VectorNode(dim)
    yet_another_node = VectorNode(dim)

    multilist.append(new_node, 0)
    multilist.append(another_node, 0)
    multilist.append(yet_another_node, 0)

    gen = multilist.reverse_iterate(0)
    assert next(gen) is yet_another_node
    assert next(gen) is another_node
    assert next(gen) is new_node

    gen = multilist.reverse_iterate(0, start=another_node)
    assert next(gen) is another_node
    assert next(gen) is new_node


def test_update_coordinate_bounds() -> None:
    bounds = np.array([-1.0, -1.0, -1.0])
    node = VectorNode(3, coordinates=[1.0, -2.0, -1.0])
    bounds = VectorLinkedList.update_coordinate_bounds(bounds, node, 0 + 1)
    assert list(bounds) == [-1, -1, -1]
    bounds = VectorLinkedList.update_coordinate_bounds(bounds, node, 1 + 1)
    assert list(bounds) == [-1, -2, -1]
    bounds = VectorLinkedList.update_coordinate_bounds(bounds, node, 2 + 1)
    assert list(bounds) == [-1, -2, -1]


def test_sort_by_index() -> None:
    nodes = [VectorNode(3, [1, 2, 3]), VectorNode(3, [2, 3, 1]), VectorNode(3, [3, 1, 2])]
    new_nodes = VectorLinkedList.sort_by_index(nodes, 0)
    assert new_nodes == nodes

    new_nodes = VectorLinkedList.sort_by_index(nodes, 1)
    assert new_nodes == [nodes[2], nodes[0], nodes[1]]

    new_nodes = VectorLinkedList.sort_by_index(nodes, 2)
    assert new_nodes == [nodes[1], nodes[2], nodes[0]]


def test_create_sorted() -> None:
    dimension = 3
    coordinates = [[1, 2, 3], [2, 3, 1], [3, 1, 2]]
    linked_list = VectorLinkedList.create_sorted(dimension, coordinates)
    assert isinstance(linked_list, VectorLinkedList)
    assert list(linked_list.sentinel.next[0].coordinates) == [1, 2, 3]
    assert list(linked_list.sentinel.next[1].coordinates) == [3, 1, 2]
    assert list(linked_list.sentinel.next[2].coordinates) == [2, 3, 1]

    assert list(linked_list.sentinel.next[0].next[0].coordinates) == [2, 3, 1]
    assert list(linked_list.sentinel.next[1].next[1].coordinates) == [1, 2, 3]
    assert list(linked_list.sentinel.next[2].next[2].coordinates) == [3, 1, 2]


def test_version_consistency() -> None:
    reference = np.array([79, 89, 99])
    hv = HypervolumeIndicator(reference)
    front = np.array(
        [(110, 110, 100), (110, 90, 87), (80, 80, 36), (50, 50, 55), (105, 30, 43), (110, 110, 100)]
    )
    volume = hv.compute(front)
    assert volume == 11113.0


def test_reference_no_pointy() -> None:
    reference = np.array([10, 10])
    hv = HypervolumeIndicator(reference)
    front = np.array(
        [
            (11, 9),
            (9, 11),
        ]
    )
    volume = hv.compute(front)
    assert volume == -3  # not sure this is expected



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from unittest import SkipTest
import pytest
import numpy as np
from nevergrad.optimization.optimizerlib import OnePlusOne
from . import core


@pytest.mark.parametrize("fid", range(1, 24))  # type: ignore
@pytest.mark.parametrize("iid", range(1, 6))  # type: ignore
def test_PBO(fid: int, iid: int) -> None:
    values = []
    try:
        func = core.PBOFunction(fid, iid, 16)
    except ModuleNotFoundError as e:
        raise SkipTest("IOH is not installed") from e
    for _ in range(30):
        x = func.parametrization.sample()
        assert not set(x.value) - {0, 1}, f"Non binary sample {x}."
        value = func(x.value)
        assert isinstance(value, float), "All output of the iohprofiler-functions should be float"
        assert np.isfinite(value)
        values.append(value)
    optim = OnePlusOne(func.parametrization, budget=100)
    recom = optim.minimize(func)
    values.append(func(recom.value))  # type: ignore
    assert (
        fid in [19, 20, 21, 22, 23] or min(values) >= 0.0 or max(values) <= 0.0
    ), f"IOH profile functions should have constant sign: pb with fid={fid},iid={iid}."


@pytest.mark.parametrize("instrumentation", ["Softmax", "Ordered"])  # type: ignore
def test_PBO_parameterization(instrumentation: str) -> None:
    try:
        func = core.PBOFunction(1, 0, 16, instrumentation=instrumentation)
    except ModuleNotFoundError as e:
        raise SkipTest("IOH is not installed") from e
    x = func.parametrization.sample()
    value = func(x.value)
    assert isinstance(value, float), "All output of the iohprofiler-functions should be float"
    assert np.isfinite(value)


def test_W_model() -> None:
    try:
        func = core.WModelFunction()
    except ModuleNotFoundError as e:
        raise SkipTest("IOH is not installed") from e
    x = func.parametrization.sample()
    func2 = core.PBOFunction(1, 0, 16)
    assert func(x.value) == func2(x.value), "W-model with default setting should equal base_function"
    func = core.WModelFunction(base_function="LeadingOnes")
    x = func.parametrization.sample()
    func2 = core.PBOFunction(2, 0, 16)
    assert func(x.value) == func2(x.value), "W-model with default setting should equal base_function"



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest


# base classes


class NevergradError(Exception):
    """Base class for error raised by Nevergrad"""


class NevergradWarning(Warning):
    pass


# errors
# pylint: disable=too-many-ancestors


class NevergradEarlyStopping(StopIteration, NevergradError):
    """Stops the minimization loop if raised"""


class NevergradRuntimeError(RuntimeError, NevergradError):
    """Runtime error raised by Nevergrad"""


class NevergradTypeError(TypeError, NevergradError):
    """Runtime error raised by Nevergrad"""


class NevergradValueError(ValueError, NevergradError):
    """Runtime error raised by Nevergrad"""


class NevergradNotImplementedError(NotImplementedError, NevergradError):
    """Not implemented functionality"""


class TellNotAskedNotSupportedError(NevergradNotImplementedError):
    """To be raised by optimizers which do not support the tell_not_asked interface."""


class ExperimentFunctionCopyError(NevergradNotImplementedError):
    """Raised when the experiment function fails to copy itself (for benchmarks)"""


class UnsupportedExperiment(unittest.SkipTest, NevergradRuntimeError):
    """Raised if the experiment is not compatible with the current settings:
    Eg: missing data, missing import, unsupported OS etc
    This automatically skips tests.
    """


class NevergradDeprecationError(NevergradRuntimeError):
    """Deprecated function/class"""


class UnsupportedParameterOperationError(NevergradRuntimeError):
    """This type of operation is not supported by the parameter"""


# warnings


class NevergradDeprecationWarning(DeprecationWarning, NevergradWarning):
    """Deprecated function/class"""


class NevergradRuntimeWarning(RuntimeWarning, NevergradWarning):
    """Runtime warning raise by nevergrad"""


class InefficientSettingsWarning(NevergradRuntimeWarning):
    """Optimization settings are not optimal for the optimizer"""


class BadLossWarning(NevergradRuntimeWarning):
    """Provided loss is unhelpful"""


class LossTooLargeWarning(BadLossWarning):
    """Sent when Loss is clipped because it is too large"""


class NevergradBehaviorChangesWarning(NevergradRuntimeWarning):
    """Notifies a change of behavior"""


class FinishedUnderlyingOptimizerWarning(NevergradRuntimeWarning):
    """Underlying scipy optimizer finished"""


class FailedConstraintWarning(NevergradRuntimeWarning):
    """Constraint could not be applied"""



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math
import operator
import numpy as np
from nevergrad.parametrization import parameter as p
from nevergrad.parametrization.utils import float_penalty as _float_penalty
import nevergrad.common.typing as tp
from nevergrad.common.tools import OrderedSet


class MultiValue:
    """Estimation of a value based on one or multiple evaluations.
    This class provides easy access to:
    - count: how many times the point was evaluated
    - mean: the mean value.
    - square: the mean square value
    - variance: the variance
    - parameter: the corresponding Parameter


    It also provides access to optimistic and pessimistic bounds for the value.

    Parameter
    ---------
    parameter: Parameter
        the parameter for one of the evaluations
    y: float
        the first evaluation of the value
    """

    def __init__(self, parameter: p.Parameter, y: float, *, reference: p.Parameter) -> None:
        self.count = 1
        self.mean = y
        self._minimum = y
        self.square = y * y
        # TODO May be safer to use a default variance which depends on y for scale invariance?
        self.variance = 1.0e6
        parameter.freeze()
        self.parameter = parameter
        self._ref = reference

    @property
    def x(self) -> np.ndarray:  # for compatibility
        return self.parameter.get_standardized_data(reference=self._ref)

    @property
    def optimistic_confidence_bound(self) -> float:
        return float(self.mean - 0.1 * np.sqrt((self.variance) / (1 + self.count)))

    @property
    def pessimistic_confidence_bound(self) -> float:
        return float(self.mean + 0.1 * np.sqrt((self.variance) / (1 + self.count)))

    def get_estimation(self, name: str) -> float:
        # Note: pruning below relies on the fact than only 3 modes exist. If a new mode is added, update pruning
        if name == "optimistic":
            return self.optimistic_confidence_bound
        elif name == "pessimistic":
            return self.pessimistic_confidence_bound
        elif name == "minimum":
            return self._minimum
        elif name == "average":
            return self.mean
        else:
            raise NotImplementedError

    def add_evaluation(self, y: float) -> None:
        """Adds a new evaluation of the value

        Parameter
        ---------
        y: float
            the new evaluation
        """
        self._minimum = min(self._minimum, y)
        self.mean = (self.count * self.mean + y) / float(self.count + 1)
        self.square = (self.count * self.square + y * y) / float(self.count + 1)
        self.square = max(self.square, self.mean**2)
        self.count += 1
        factor = math.sqrt(float(self.count) / float(self.count - 1.0))
        self.variance = factor * (self.square - self.mean**2)

    def as_array(self, reference: p.Parameter) -> np.ndarray:
        return self.parameter.get_standardized_data(reference=reference)

    def __repr__(self) -> str:
        return f"MultiValue<mean: {self.mean}, count: {self.count}, parameter: {self.parameter}>"


def _get_nash(optimizer: tp.Any) -> tp.List[tp.Tuple[tp.Tuple[float, ...], int]]:
    """Returns an empirical distribution. limited using a threshold
    equal to max_num_trials^(1/4).
    """
    if not optimizer.archive:
        return [(optimizer.current_bests["pessimistic"].x, 1)]
    max_num_trial = max(p.count for p in optimizer.archive.values())
    sum_num_trial = sum(p.count for p in optimizer.archive.values())
    threshold = np.power(max_num_trial, 0.5)
    if threshold <= np.power(sum_num_trial, 0.25):
        return [(optimizer.provide_recommendation(), 1)]
    # make deterministic at the price of sort complexity
    return sorted(
        ((np.frombuffer(k), p.count) for k, p in optimizer.archive.bytesdict.items() if p.count >= threshold),  # type: ignore
        key=operator.itemgetter(1),
    )


def sample_nash(optimizer: tp.Any) -> tp.Tuple[float, ...]:  # Somehow like fictitious play.
    nash = _get_nash(optimizer)
    if len(nash) == 1:
        return nash[0][0]
    prob = [float(n[1]) for n in nash]
    prob = [p_ / sum(prob) for p_ in prob]
    index: int = np.random.choice(np.arange(len(prob)), p=prob)
    return nash[index][0]


class DelayedJob:
    """Future-like object which delays computation"""

    def __init__(self, func: tp.Callable[..., tp.Any], *args: tp.Any, **kwargs: tp.Any) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self._result: tp.Optional[tp.Any] = None
        self._computed = False

    def done(self) -> bool:
        return True

    def result(self) -> tp.Any:
        if not self._computed:
            self._result = self.func(*self.args, **self.kwargs)
            self._computed = True
        return self._result


class SequentialExecutor:
    """Executor which run sequentially and locally
    (just calls the function and returns a FinishedJob)
    """

    def submit(self, fn: tp.Callable[..., tp.Any], *args: tp.Any, **kwargs: tp.Any) -> DelayedJob:
        return DelayedJob(fn, *args, **kwargs)


def _tobytes(x: tp.ArrayLike) -> bytes:
    x = np.array(x, copy=False)  # for compatibility
    assert x.ndim == 1, f"Input shape: {x.shape}"
    assert x.dtype == np.float_, f"Incorrect type {x.dtype} is not float"
    return x.tobytes()


_ERROR_STR = (
    "Generating numpy arrays from the bytes keys is inefficient, "
    "work on archive.bytesdict.<keys,items>() directly and convert with "
    "np.frombuffer if you can. You can also use archive.<keys,items>_as_arrays() "
    "but it is less efficient."
)


Y = tp.TypeVar("Y")


class Archive(tp.Generic[Y]):
    """A dict-like object with numpy arrays as keys.
    The underlying `bytesdict` dict stores the arrays as bytes since arrays are not hashable.
    Keys can be converted back with np.frombuffer(key)
    """

    def __init__(self) -> None:
        self.bytesdict: tp.Dict[bytes, Y] = {}

    def __setitem__(self, x: tp.ArrayLike, value: Y) -> None:
        self.bytesdict[_tobytes(x)] = value

    def __getitem__(self, x: tp.ArrayLike) -> Y:
        return self.bytesdict[_tobytes(x)]

    def __contains__(self, x: tp.ArrayLike) -> bool:
        return _tobytes(x) in self.bytesdict

    def get(self, x: tp.ArrayLike, default: tp.Optional[Y] = None) -> tp.Optional[Y]:
        return self.bytesdict.get(_tobytes(x), default)

    def __len__(self) -> int:
        return len(self.bytesdict)

    def values(self) -> tp.ValuesView[Y]:
        return self.bytesdict.values()

    def keys(self) -> None:
        raise RuntimeError(_ERROR_STR)

    def items(self) -> None:
        raise RuntimeError(_ERROR_STR)

    def items_as_array(self) -> tp.Iterator[tp.Tuple[np.ndarray, Y]]:
        raise RuntimeError("For consistency, items_as_array is renamed to items_as_arrays")

    def items_as_arrays(self) -> tp.Iterator[tp.Tuple[np.ndarray, Y]]:
        """Functions that iterates on key-values but transforms keys
        to np.ndarray. This is to simplify interactions, but should not
        be used in an algorithm since the conversion can be inefficient.
        Prefer using self.bytesdict.items() directly, and convert the bytes
        to np.ndarray using np.frombuffer(b)
        """
        return ((np.frombuffer(b), v) for b, v in self.bytesdict.items())

    def keys_as_array(self) -> tp.Iterator[np.ndarray]:
        raise RuntimeError("For consistency, keys_as_array is renamed to keys_as_arrays")

    def keys_as_arrays(self) -> tp.Iterator[np.ndarray]:
        """Functions that iterates on keys but transforms them
        to np.ndarray. This is to simplify interactions, but should not
        be used in an algorithm since the conversion can be inefficient.
        Prefer using self.bytesdict.keys() directly, and convert the bytes
        to np.ndarray using np.frombuffer(b)
        """
        return (np.frombuffer(b) for b in self.bytesdict)

    def __repr__(self) -> str:
        return f"Archive with bytesdict: {self.bytesdict!r}"

    def __str__(self) -> str:
        return f"Archive with bytesdict: {self.bytesdict}"

    def __iter__(self) -> None:
        raise RuntimeError(_ERROR_STR)


class Pruning:
    """Callable for pruning archives in the optimizer class.
    See Optimizer.pruning attribute, called at each "tell".

    Parameters
    ----------
    min_len: int
        minimum length of the pruned archive.
    max_len: int
        length at which pruning is activated (maximum allowed length for the archive).

    Note
    ----
    For each of the 3 criteria (optimistic, pessimistic and average), the min_len best (lowest)
    points will be kept, which can lead to at most 3 * min_len points.
    """

    def __init__(self, min_len: int, max_len: int):
        self.min_len = min_len
        self.max_len = max_len
        self._num_prunings = 0  # for testing it is not called too often

    def __call__(self, archive: Archive[MultiValue]) -> Archive[MultiValue]:
        if len(archive) < self.max_len:
            return archive
        return self._prune(archive)

    def _prune(self, archive: Archive[MultiValue]) -> Archive[MultiValue]:
        self._num_prunings += 1
        # separate function to ease profiling
        quantiles: tp.Dict[str, float] = {}
        threshold = float(self.min_len + 1) / len(archive)
        names = ["optimistic", "pessimistic", "average"]
        for name in names:
            quantiles[name] = np.quantile(  # type: ignore
                [v.get_estimation(name) for v in archive.values()], threshold, method="lower"
            )
        new_archive: Archive[MultiValue] = Archive()
        new_archive.bytesdict = {
            b: v
            for b, v in archive.bytesdict.items()
            if any(v.get_estimation(n) < quantiles[n] for n in names)
        }  # strict comparison to make sure we prune even for values repeated maaany times
        # this may remove all points though, but nevermind for now
        return new_archive

    @classmethod
    def sensible_default(cls, num_workers: int, dimension: int) -> "Pruning":
        """Very conservative pruning
        - keep at least 100 elements, or 7 times num_workers, whatever is biggest
        - keep at least 3 x min_len, or up to 10 x min_len if it does not exceed 1gb of data

        Parameters
        ----------
        num_workers: int
            number of evaluations which will be run in parallel at once
        dimension: int
            dimension of the optimization space
        """
        # safer to keep at least 7 time the workers
        min_len = max(100, 7 * num_workers)
        max_len_1gb = 1024**3 // (dimension * 8 * 2)  # stored twice: as key and as Parameter
        max_len = max(3 * min_len, min(10 * min_len, max_len_1gb))
        return cls(min_len, max_len)


class UidQueue:
    """Queue of uids to handle a population. This keeps track of:
    - told uids
    - asked uids
    When telling, it removes from the asked queue and adds to the told queue
    When asking, it takes from the told queue if not empty, else from the older
    asked, and then adds to the asked queue.
    """

    def __init__(self) -> None:
        self.told = tp.Deque[str]()  # this seems to be picklable (this syntax does not always work)
        self.asked: OrderedSet[str] = OrderedSet()

    def clear(self) -> None:
        """Removes all uids from the queues"""
        self.told.clear()
        self.asked.clear()

    def ask(self) -> str:
        """Takes a uid from the told queue if not empty, else from the older asked,
        then adds it to the asked queue.
        """
        if self.told:
            uid = self.told.popleft()
        elif self.asked:
            uid = next(iter(self.asked))
        else:
            raise RuntimeError("Both asked and told queues are empty.")
        self.asked.add(uid)
        return uid

    def tell(self, uid: str) -> None:
        """Removes the uid from the asked queue and adds to the told queue"""
        self.told.append(uid)
        if uid in self.asked:
            self.asked.discard(uid)

    def discard(self, uid: str) -> None:
        if uid in self.asked:
            self.asked.discard(uid)
        else:
            self.told.remove(uid)


class ConstraintManager:
    """Try max_constraints_trials random explorations for satisfying constraints.
    The finally chosen point, if it does not satisfy constraints, is penalized as shown in the penalty function,
    using coeffcieints mentioned here.


    Possibly unstable.
    """

    def __init__(self) -> None:
        self.max_trials = 1000
        self.penalty_factor = 1.0
        self.penalty_exponent = 1.001

    def __repr__(self) -> str:
        return "Constraints:" + ",".join(f"{x}={y}" for x, y in self.__dict__.items())

    # pylint: disable=unused-argument
    def update(
        self,
        max_trials: tp.Optional[int] = None,
        penalty_factor: tp.Optional[float] = None,
        penalty_exponent: tp.Optional[float] = None,
    ) -> None:
        """
        Parameters
        ----------
            max_trials: int
                number of random tries for satisfying constraints.
            penalty: float
                multiplicative factor on the constraint penalization.
            penalty_exponent: float
                exponent, usually close to 1 and slightly greater than 1.
        """
        for x, y in locals().items():
            if y is not None and x != "self" and not x.startswith("_"):
                setattr(self, x, y)
        if self.penalty_exponent < 1:
            raise ValueError("Penalty exponent needs to be equal or greater than 1")

    def penalty(self, parameter: p.Parameter, num_ask: int, budget: tp.Optional[int]) -> float:
        """Computes the penalty associated with a Parameter, for constraint management"""
        budget = 1 if budget is None else budget
        coeff = self.penalty_factor * (self.penalty_exponent ** (num_ask / np.sqrt(budget)))
        val = parameter.value
        return coeff * sum(_float_penalty(func(val)) for func in parameter._constraint_checkers)  # type: ignore



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .common import typing as typing
from .parametrization import parameter as p
from .optimization import optimizerlib as optimizers  # busy namespace, likely to be simplified
from .optimization import families as families
from .optimization import callbacks as callbacks
from .common import errors as errors
from . import ops as ops


__all__ = ["optimizers", "families", "callbacks", "p", "typing", "errors", "ops"]


__version__ = "1.0.0"



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from . import envs as envs
from . import agents as agents
from .base import EnvironmentRunner as EnvironmentRunner



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
from . import base
from .base import IntOrParameter
import nevergrad.common.typing as tp
from nevergrad.parametrization import transforms
from nevergrad.parametrization import parameter as p

import hyperopt  # type: ignore
from hyperopt import hp, Trials, Domain, tpe  # type: ignore


def _hp_parametrization_to_dict(x, **kwargs):
    if isinstance(x, p.Instrumentation):
        x_dict = kwargs.get("default", {})
        for idx_param in range(len(x[0].value)):
            x_dict.update(_hp_parametrization_to_dict(x[0][idx_param], name=str(idx_param)))
        for name in x[1].value.keys():
            x_dict.update(_hp_parametrization_to_dict(x[1][name], name=name))
        return x_dict
    elif isinstance(x, (p.Log, p.Scalar)):
        return {kwargs["name"]: [x.value]}
    elif isinstance(x, p.Choice):
        x_dict = {}
        for i in range(len(x.choices)):
            if x.value == x.choices[i].value:
                x_dict[kwargs["name"]] = [i]
                if isinstance(x.choices[i], (p.Log, p.Scalar, p.Choice)):
                    x_dict[kwargs["name"] + f"__{i}"] = [x.choices[i].value]
                elif isinstance(x.choices[i], p.Instrumentation):
                    x_dict.update(_hp_parametrization_to_dict(x.choices[i]))
        return x_dict


def _hp_dict_to_parametrization(x):
    if isinstance(x, dict) and ("args" in x) and ("kwargs" in x):
        x["args"] = tuple([_hp_dict_to_parametrization(x["args"][str(i)]) for i in range(len(x["args"]))])
        x["kwargs"] = {k: _hp_dict_to_parametrization(v) for k, v in x["kwargs"].items()}
        return (x["args"], x["kwargs"])
    return x


def _get_search_space(param_name, param):
    if isinstance(param, p.Instrumentation):
        space = {}
        space["args"] = {
            str(idx_param): _get_search_space(str(idx_param), param[0][idx_param])  # type: ignore
            for idx_param in range(len(param[0].value))
        }
        space["kwargs"] = {
            param_name: _get_search_space(param_name, param[1][param_name])  # type: ignore
            for param_name in param[1].value.keys()
        }
        return space

    elif isinstance(param, (p.Log, p.Scalar)):
        if (param.bounds[0][0] is None) or (param.bounds[1][0] is None):
            if isinstance(param, p.Scalar) and not param.integer:
                return hp.lognormal(label=param_name, mu=0, sigma=1)
            raise ValueError(f"Scalar {param_name} not bounded.")
        elif isinstance(param, p.Log):
            return hp.loguniform(
                label=param_name, low=np.log(param.bounds[0][0]), high=np.log(param.bounds[1][0])
            )
        elif isinstance(param, p.Scalar):
            if param.integer:
                return hp.randint(label=param_name, low=int(param.bounds[0][0]), high=int(param.bounds[1][0]))
            else:
                return hp.uniform(label=param_name, low=param.bounds[0][0], high=param.bounds[1][0])

    elif isinstance(param, p.Choice):
        list_types = [
            type(param.choices[i])
            for i in range(len(param.choices))
            if not isinstance(param.choices[i], (p.Instrumentation, p.Constant))
        ]

        if len(list_types) != len(set(list_types)):
            raise NotImplementedError
        return hp.choice(
            param_name,
            [
                _get_search_space(param_name + "__" + str(i), param.choices[i])
                for i in range(len(param.choices))
            ],
        )

    elif isinstance(param, p.Constant):
        return param.value

    # Hyperopt do not support array
    raise NotImplementedError


class _HyperOpt(base.Optimizer):
    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        parametrization: IntOrParameter,
        budget: tp.Optional[int] = None,
        num_workers: int = 1,
        *,
        prior_weight: float = 1.0,
        n_startup_jobs: int = 20,
        n_EI_candidates: int = 24,
        gamma: float = 0.25,
        verbose: bool = False,
    ) -> None:
        super().__init__(parametrization, budget=budget, num_workers=num_workers)
        try:
            # try to convert parametrization to hyperopt search space
            if not isinstance(self.parametrization, p.Instrumentation):
                raise NotImplementedError
            self.space = _get_search_space(self.parametrization.name, self.parametrization)
            self._transform = None
        except NotImplementedError:
            self._transform = transforms.ArctanBound(0, 1)
            self.space = {f"x_{i}": hp.uniform(f"x_{i}", 0, 1) for i in range(self.dimension)}

        self.trials = Trials()
        self.domain = Domain(fn=None, expr=self.space, pass_expr_memo_ctrl=False)
        self.tpe_args = {
            "prior_weight": prior_weight,
            "n_startup_jobs": n_startup_jobs,
            "n_EI_candidates": n_EI_candidates,
            "gamma": gamma,
            "verbose": verbose,
        }

    def _internal_ask_candidate(self) -> p.Parameter:
        # Inspired from FMinIter class (hyperopt)
        next_id = self.trials.new_trial_ids(1)
        new_trial = tpe.suggest(
            next_id, self.domain, self.trials, self._rng.randint(2**31 - 1), **self.tpe_args
        )[0]
        self.trials.insert_trial_doc(new_trial)
        self.trials.refresh()

        candidate = self.parametrization.spawn_child()

        if self._transform:
            data = np.array([new_trial["misc"]["vals"][f"x_{i}"][0] for i in range(self.dimension)])
            candidate = candidate.set_standardized_data(self._transform.backward(data))

            # For consistency, we need to update hyperopt history
            # when standardized data is changed
            if any(
                data
                != self._transform.forward(candidate.get_standardized_data(reference=self.parametrization))
            ):
                for it, val in enumerate(
                    self._transform.forward(candidate.get_standardized_data(reference=self.parametrization))
                ):
                    self.trials._dynamic_trials[next_id[0]]["misc"]["vals"][f"x_{it}"][0] = val
        else:
            spec = hyperopt.base.spec_from_misc(new_trial["misc"])
            config = hyperopt.space_eval(self.space, spec)
            candidate.value = _hp_dict_to_parametrization(config)

        candidate._meta["trial_id"] = new_trial["tid"]
        return candidate

    def _internal_tell_candidate(self, candidate: p.Parameter, loss: float) -> None:
        result = {"loss": loss, "status": "ok"}
        assert "trial_id" in candidate._meta
        tid = candidate._meta["trial_id"]
        assert self.trials._dynamic_trials[tid]["state"] == hyperopt.JOB_STATE_NEW

        now = hyperopt.utils.coarse_utcnow()
        self.trials._dynamic_trials[tid]["book_time"] = now
        self.trials._dynamic_trials[tid]["refresh_time"] = now
        self.trials._dynamic_trials[tid]["state"] = hyperopt.JOB_STATE_DONE
        self.trials._dynamic_trials[tid]["result"] = result
        self.trials._dynamic_trials[tid]["refresh_time"] = hyperopt.utils.coarse_utcnow()
        self.trials.refresh()

    def _internal_tell_not_asked(self, candidate: p.Parameter, loss: float) -> None:
        next_id = self.trials.new_trial_ids(1)
        new_trial = hyperopt.rand.suggest(next_id, self.domain, self.trials, self._rng.randint(2**31 - 1))
        self.trials.insert_trial_docs(new_trial)
        self.trials.refresh()
        tid = next_id[0]

        if self._transform:
            data = candidate.get_standardized_data(reference=self.parametrization)
            data = self._transform.forward(data)
            self.trials._dynamic_trials[tid]["misc"]["vals"] = {f"x_{i}": [data[i]] for i in range(len(data))}
        else:
            null_config: dict = {k: [] for k in self.trials._dynamic_trials[tid]["misc"]["vals"].keys()}
            new_vals: dict = _hp_parametrization_to_dict(candidate, default=null_config)
            self.trials._dynamic_trials[tid]["misc"]["vals"] = new_vals

        self.trials.refresh()
        candidate._meta["trial_id"] = tid
        self._internal_tell_candidate(candidate, loss)


class ParametrizedHyperOpt(base.ConfiguredOptimizer):
    # pylint: disable=too-many-instance-attributes
    """Hyperopt: Distributed Asynchronous Hyper-parameter Optimization.
    This class is a wrapper over the `hyperopt <https://github.com/hyperopt/hyperopt>`_ package.

    Parameters
    ----------
    parametrization: int or Parameter
        Parametrization object
    budget: int
        Number of iterations
    num_workers: int
        Number of workers
    prior_weight: float (default 1.0)
        Smoothing factor to avoid having zero probabilities
    n_startup_jobs: int (default 20)
        Number of random uniform suggestions at initialization
    n_EI_candidates: int (default 24)
        Number of generated candidates during EI maximization
    gamma: float (default 0.25)
        Threshold to split between l(x) and g(x), see eq. 2 in

    verbose: bool (default False)
        Hyperopt algorithm verbosity

    Note
    ----
    HyperOpt is described in Bergstra, James S., et al.
    "Algorithms for hyper-parameter optimization."
    Advances in neural information processing systems. 2011
    """

    no_parallelization = False

    # pylint: disable=unused-argument
    def __init__(
        self,
        *,
        prior_weight: float = 1.0,
        n_startup_jobs: int = 20,
        n_EI_candidates: int = 24,
        gamma: float = 0.25,
        verbose: bool = False,
    ) -> None:
        super().__init__(_HyperOpt, locals())


HyperOpt = ParametrizedHyperOpt().set_name("HyperOpt", register=True)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from nevergrad.parametrization.instantiate import FolderFunction as FolderFunction
from nevergrad.parametrization.utils import TemporaryDirectoryCopy as TemporaryDirectoryCopy
from nevergrad.parametrization.utils import CommandFunction as CommandFunction
from . import parameter as parameter



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pylint: disable=unused-import
# import with "as" to explicitely allow reexport (mypy)

# abstract types
from .core import Parameter as Parameter
from .container import Container as Container
from .choice import BaseChoice as BaseChoice
from .data import Data as Data

# special types
from .core import Constant as Constant  # avoid using except for checks
from .core import MultiobjectiveReference as MultiobjectiveReference  # multiobjective optimization

# containers
from .container import Dict as Dict
from .container import Tuple as Tuple
from .container import Instrumentation as Instrumentation

# data
from .data import Array as Array
from .data import Scalar as Scalar
from .data import Log as Log
from . import mutation
from ._datalayers import Angles as Angles

# choices
from .choice import Choice as Choice
from .choice import TransitionChoice as TransitionChoice

# other
from . import helpers as helpers



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from collections import abc
import numpy as np
import pandas as pd
import nevergrad.common.typing as tp
from nevergrad.common import testing


class Selector(pd.DataFrame):  # type: ignore
    """Pandas dataframe class with a simplified selection function"""

    @property
    def _constructor_expanddim(self) -> tp.Type["Selector"]:
        return Selector

    @property
    def _constructor(self) -> tp.Type["Selector"]:
        return Selector

    # pylint: disable=arguments-differ
    def select(self, **kwargs: tp.Union[str, tp.Sequence[str], tp.Callable[[tp.Any], bool]]) -> "Selector":
        """Select rows based on a value, a sequence of values or a discriminating function

        Parameters
        ----------
        kwargs: str, list or callable
            selects values in the column provided as keyword, based on str matching, or
            presence in the list, or callable returning non-False on the values

        Example
        -------
        df.select(column1=["a", "b"])
        will return a new Selector with rows having either "a" or "b" as value in column1
        """
        df = self
        for name, criterion in kwargs.items():
            if isinstance(criterion, abc.Iterable) and not isinstance(criterion, str):
                selected = df.loc[:, name].isin(criterion)
            elif callable(criterion):
                selected = [bool(criterion(x)) for x in df.loc[:, name]]
            else:
                selected = df.loc[:, name].isin([criterion])
            df = df.loc[selected, :]
        return Selector(df)

    def select_and_drop(
        self, **kwargs: tp.Union[str, tp.Sequence[str], tp.Callable[[tp.Any], bool]]
    ) -> "Selector":
        """Same as select, but drops the columns used for selection"""
        df = self.select(**kwargs)
        columns = [x for x in df.columns if x not in kwargs]
        return Selector(df.loc[:, columns])

    def unique(
        self, column_s: tp.Union[str, tp.Sequence[str]]
    ) -> tp.Union[tp.Tuple[tp.Any, ...], tp.Set[tp.Tuple[tp.Any, ...]]]:
        """Returns the set of unique values or set of values for a column or columns

        Parameter
        ---------
        column_s: str or tp.Sequence[str]
            a column name, or list of column names

        Returns
        -------
        set
           a set of values if the input was a column name, or a set of tuple of values
           if the name was a list of columns
        """
        if isinstance(column_s, str):
            return set(self.loc[:, column_s])  # equivalent to df.<name>.unique()
        elif isinstance(column_s, (list, tuple)):
            testing.assert_set_equal(set(column_s) - set(self.columns), {}, err_msg="Unknown column(s)")
            df = self.loc[:, column_s]
            assert not df.isnull().values.any(), "Cannot work with NaN values"
            return set(tuple(row) for row in df.itertuples(index=False))
        else:
            raise NotImplementedError("Only strings, lists and tuples are allowed")

    @classmethod
    def read_csv(cls, path: tp.PathLike) -> "Selector":
        return cls(pd.read_csv(str(path)))

    def assert_equivalent(self, other: pd.DataFrame, err_msg: str = "") -> None:
        """Asserts that two selectors are equal, up to row and column permutations

        Note
        ----
        Use sparsely, since it is quite slow to test
        """
        testing.assert_set_equal(other.columns, self.columns, f"Different columns\n{err_msg}")
        np.testing.assert_equal(len(other), len(self), "Different number of rows\n{err_msg}")
        other_df = other.loc[:, self.columns]
        df_rows: tp.List[tp.List[tp.Tuple[tp.Any, ...]]] = [[], []]
        for k, df in enumerate([self, other_df]):
            for row in df.itertuples(index=False):
                df_rows[k].append(tuple(row))
            df_rows[k].sort()
        for row1, row2 in zip(*df_rows):
            np.testing.assert_array_equal(row1, row2, err_msg=err_msg)



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import nevergrad.common.typing as tp
from . import utils
from .base import registry
from . import callbacks


class MetaModelFailure(ValueError):
    """Sometimes the optimum of the metamodel is at infinity."""


def learn_on_k_best(
    archive: utils.Archive[utils.MultiValue], k: int, algorithm: str = "quad"
) -> tp.ArrayLike:
    """Approximate optimum learnt from the k best.

    Parameters
    ----------
    archive: utils.Archive[utils.Value]
    """
    items = list(archive.items_as_arrays())
    dimension = len(items[0][0])

    # Select the k best.
    first_k_individuals = sorted(items, key=lambda indiv: archive[indiv[0]].get_estimation("average"))[:k]
    assert len(first_k_individuals) == k

    # Recenter the best.
    middle = np.array(sum(p[0] for p in first_k_individuals) / k)
    normalization = 1e-15 + np.sqrt(np.sum((first_k_individuals[-1][0] - first_k_individuals[0][0]) ** 2))
    y = np.asarray([archive[c[0]].get_estimation("pessimistic") for c in first_k_individuals])
    X = np.asarray([(c[0] - middle) / normalization for c in first_k_individuals])

    from sklearn.preprocessing import PolynomialFeatures

    polynomial_features = PolynomialFeatures(degree=2)
    X2 = polynomial_features.fit_transform(X)
    if not max(y) - min(y) > 1e-20:  # better use "not" for dealing with nans
        raise MetaModelFailure
    y = (y - min(y)) / (max(y) - min(y))
    if algorithm == "neural":
        from sklearn.neural_network import MLPRegressor

        model = MLPRegressor(hidden_layer_sizes=(16, 16), solver="lbfgs")
    elif algorithm in ["svm", "svr"]:
        from sklearn.svm import SVR

        model = SVR()
    elif algorithm == "rf":
        from sklearn.ensemble import RandomForestRegressor

        model = RandomForestRegressor()
    else:
        assert algorithm == "quad", f"Metamodelling algorithm {algorithm} not recognized."
        # We need SKLearn.
        from sklearn.linear_model import LinearRegression

        # Fit a linear model.
        model = LinearRegression()

    model.fit(X2, y)
    # Check model quality.
    model_outputs = model.predict(X2)
    indices = np.argsort(y)
    ordered_model_outputs = [model_outputs[i] for i in indices]
    if not np.all(np.diff(ordered_model_outputs) > 0):
        raise MetaModelFailure("Unlearnable objective function.")

    try:
        Powell = registry["Powell"]
        DE = registry["DE"]
        for cls in (Powell, DE):  # Powell excellent here, DE as a backup for thread safety.
            optimizer = cls(parametrization=dimension, budget=45 * dimension + 30)
            # limit to 20s at most
            optimizer.register_callback("ask", callbacks.EarlyStopping.timer(20))
            try:
                minimum = optimizer.minimize(
                    lambda x: float(model.predict(polynomial_features.fit_transform(x[None, :])))
                ).value
            except RuntimeError:
                assert cls == Powell, "Only Powell is allowed to crash here."
            else:
                break
    except ValueError:
        raise MetaModelFailure("Infinite meta-model optimum in learn_on_k_best.")
    if float(model.predict(polynomial_features.fit_transform(minimum[None, :]))) > y[0]:
        raise MetaModelFailure("Not a good proposal.")
    if np.sum(minimum**2) > 1.0:
        raise MetaModelFailure("huge meta-model optimum in learn_on_k_best.")
    return middle + normalization * minimum



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import uuid
import itertools
import numpy as np
from scipy import stats
import nevergrad.common.typing as tp
from . import utils


def bound_to_array(x: tp.BoundValue) -> np.ndarray:
    """Updates type of bounds to use arrays"""
    if isinstance(x, (tuple, list, np.ndarray)):
        return np.array(x, copy=False)
    else:
        return np.array([x], dtype=float)


class Transform:
    """Base class for transforms implementing a forward and a backward (inverse)
    method.
    This provide a default representation, and a short representation should be implemented
    for each transform.
    """

    def __init__(self) -> None:
        self.name = uuid.uuid4().hex  # a name for easy identification. This random uuid should be overriden

    def forward(self, x: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    def backward(self, y: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    def reverted(self) -> "Transform":
        return Reverted(self)

    def __repr__(self) -> str:
        args = ", ".join(f"{x}={y}" for x, y in sorted(self.__dict__.items()) if not x.startswith("_"))
        return f"{self.__class__.__name__}({args})"


class Reverted(Transform):
    """Inverse of a transform.

    Parameters
    ----------
    transform: Transform
    """

    def __init__(self, transform: Transform) -> None:
        super().__init__()
        self.transform = transform
        self.name = f"Rv({self.transform.name})"

    def forward(self, x: np.ndarray) -> np.ndarray:
        return self.transform.backward(x)

    def backward(self, y: np.ndarray) -> np.ndarray:
        return self.transform.forward(y)


class Affine(Transform):
    """Affine transform a * x + b

    Parameters
    ----------
    a: float
    b: float
    """

    def __init__(self, a: tp.BoundValue, b: tp.BoundValue) -> None:
        super().__init__()
        self.a = bound_to_array(a)
        self.b = bound_to_array(b)
        if not np.any(self.a):
            raise ValueError('"a" parameter should be non-zero to prevent information loss.')
        self.name = f"Af({a},{b})"

    def forward(self, x: np.ndarray) -> np.ndarray:
        return self.a * x + self.b  # type: ignore

    def backward(self, y: np.ndarray) -> np.ndarray:
        return (y - self.b) / self.a  # type: ignore


class Exponentiate(Transform):
    """Exponentiation transform base ** (coeff * x)
    This can for instance be used for to get a logarithmicly distruted values 10**(-[1, 2, 3]).

    Parameters
    ----------
    base: float
    coeff: float
    """

    def __init__(self, base: float = 10.0, coeff: float = 1.0) -> None:
        super().__init__()
        self.base = base
        self.coeff = coeff
        self.name = f"Ex({self.base},{self.coeff})"

    def forward(self, x: np.ndarray) -> np.ndarray:
        return self.base ** (float(self.coeff) * x)  # type: ignore

    def backward(self, y: np.ndarray) -> np.ndarray:
        return np.log(y) / (float(self.coeff) * np.log(self.base))  # type: ignore


BoundType = tp.Optional[tp.Union[tp.ArrayLike, float]]


def _f(x: BoundType) -> BoundType:
    """Format for prints:
    array with one scalars are converted to floats
    """
    if isinstance(x, (np.ndarray, list, tuple)):
        x = np.array(x, copy=False)
        if x.shape == (1,):
            x = float(x[0])
    if isinstance(x, float) and x.is_integer():
        x = int(x)
    return x


class BoundTransform(Transform):  # pylint: disable=abstract-method
    def __init__(self, a_min: BoundType = None, a_max: BoundType = None) -> None:
        super().__init__()
        self.a_min: tp.Optional[np.ndarray] = None
        self.a_max: tp.Optional[np.ndarray] = None
        for name, value in [("a_min", a_min), ("a_max", a_max)]:
            if value is not None:
                isarray = isinstance(value, (tuple, list, np.ndarray))
                setattr(self, name, np.array(value, copy=False) if isarray else np.array([value]))
        if not (self.a_min is None or self.a_max is None):
            if (self.a_min >= self.a_max).any():
                raise ValueError(f"Lower bounds {a_min} should be strictly smaller than upper bounds {a_max}")
        if self.a_min is None and self.a_max is None:
            raise ValueError("At least one bound must be specified")
        self.shape: tp.Tuple[int, ...] = self.a_min.shape if self.a_min is not None else self.a_max.shape

    def _check_shape(self, x: np.ndarray) -> None:
        for dims in itertools.zip_longest(x.shape, self.shape, fillvalue=1):
            if dims[0] != dims[1] and not any(x == 1 for x in dims):  # same or broadcastable
                raise ValueError(f"Shapes do not match: {self.shape} and {x.shape}")


class TanhBound(BoundTransform):
    """Bounds all real values into [a_min, a_max] using a tanh transform.
    Beware, tanh goes very fast to its limits.

    Parameters
    ----------
    a_min: float
    a_max: float
    """

    def __init__(self, a_min: tp.Union[tp.ArrayLike, float], a_max: tp.Union[tp.ArrayLike, float]) -> None:
        super().__init__(a_min=a_min, a_max=a_max)
        if self.a_min is None or self.a_max is None:
            raise ValueError("Both bounds must be specified")
        self._b = 0.5 * (self.a_max + self.a_min)
        self._a = 0.5 * (self.a_max - self.a_min)
        self.name = f"Th({_f(a_min)},{_f(a_max)})"

    def forward(self, x: np.ndarray) -> np.ndarray:
        self._check_shape(x)
        return self._b + self._a * np.tanh(x)  # type: ignore

    def backward(self, y: np.ndarray) -> np.ndarray:
        self._check_shape(y)
        if (y > self.a_max).any() or (y < self.a_min).any():
            raise ValueError(
                f"Only data between {self.a_min} and {self.a_max} "
                "can be transformed back (bounds lead to infinity)."
            )
        return np.arctanh((y - self._b) / self._a)  # type: ignore


class Clipping(BoundTransform):
    """Bounds all real values into [a_min, a_max] using clipping (not bijective).

    Parameters
    ----------
    a_min: float or None
        lower bound
    a_max: float or None
        upper bound
    bounce: bool
        bounce (once) on borders instead of just clipping
    """

    def __init__(
        self,
        a_min: BoundType = None,
        a_max: BoundType = None,
        bounce: bool = False,
    ) -> None:
        super().__init__(a_min=a_min, a_max=a_max)
        self._bounce = bounce
        b = ",b" if bounce else ""
        self.name = f"Cl({_f(a_min)},{_f(a_max)}{b})"
        self.checker = utils.BoundChecker(self.a_min, self.a_max)

    def forward(self, x: np.ndarray) -> np.ndarray:
        self._check_shape(x)
        if self.checker(x):
            return x
        out = np.clip(x, self.a_min, self.a_max)  # type: ignore
        if self._bounce:
            out = np.clip(2 * out - x, self.a_min, self.a_max)  # type: ignore
        return out  # type: ignore

    def backward(self, y: np.ndarray) -> np.ndarray:
        self._check_shape(y)
        if not self.checker(y):
            raise ValueError(
                f"Only data between {self.a_min} and {self.a_max} can be transformed back.\n" f"Got: {y}"
            )
        return y


class ArctanBound(BoundTransform):
    """Bounds all real values into [a_min, a_max] using an arctan transform.
    This is a much softer approach compared to tanh.

    Parameters
    ----------
    a_min: float
    a_max: float
    """

    def __init__(self, a_min: tp.Union[tp.ArrayLike, float], a_max: tp.Union[tp.ArrayLike, float]) -> None:
        super().__init__(a_min=a_min, a_max=a_max)
        if self.a_min is None or self.a_max is None:
            raise ValueError("Both bounds must be specified")
        self._b = 0.5 * (self.a_max + self.a_min)
        self._a = (self.a_max - self.a_min) / np.pi
        self.name = f"At({_f(a_min)},{_f(a_max)})"

    def forward(self, x: np.ndarray) -> np.ndarray:
        self._check_shape(x)
        return self._b + self._a * np.arctan(x)  # type: ignore

    def backward(self, y: np.ndarray) -> np.ndarray:
        self._check_shape(y)
        if (y > self.a_max).any() or (y < self.a_min).any():
            raise ValueError(f"Only data between {self.a_min} and {self.a_max} can be transformed back.")
        return np.tan((y - self._b) / self._a)  # type: ignore


class CumulativeDensity(BoundTransform):
    """Bounds all real values into [0, 1] using a gaussian cumulative density function (cdf)
    Beware, cdf goes very fast to its limits.

    Parameters
    ----------
    lower: float
        lower bound
    upper: float
        upper bound
    eps: float
        small values to avoid hitting the bounds
    scale: float
        scaling factor of the density
    density: str
        either gaussian, or cauchy distributions
    """

    def __init__(
        self,
        lower: float = 0.0,
        upper: float = 1.0,
        eps: float = 1e-9,
        scale: float = 1.0,
        density: str = "gaussian",
    ) -> None:
        super().__init__(a_min=lower, a_max=upper)
        self._b = lower
        self._a = upper - lower
        self._eps = eps
        self._scale = scale
        self.name = f"Cd({_f(lower)},{_f(upper)})"
        if density not in ("gaussian", "cauchy"):
            raise ValueError("Unknown density")
        if density == "gaussian":
            self._forw = stats.norm.cdf
            self._back = stats.norm.ppf
        else:
            self._forw = stats.cauchy.cdf
            self._back = stats.cauchy.ppf

    def forward(self, x: np.ndarray) -> np.ndarray:
        return self._a * self._forw(x / self._scale) + self._b  # type: ignore

    def backward(self, y: np.ndarray) -> np.ndarray:
        if (y > self.a_max).any() or (y < self.a_min).any():
            raise ValueError(
                f"Only data between {self.a_min} and {self.a_max} can be transformed back.\nGot: {y}"
            )
        y = np.clip((y - self._b) / self._a, self._eps, 1 - self._eps)
        return self._scale * self._back(y)


class Fourrier(Transform):
    def __init__(self, axes: tp.Union[int, tp.Sequence[int]] = 0) -> None:
        super().__init__()
        self.axes: tp.Tuple[int, ...] = (axes,) if isinstance(axes, int) else tuple(axes)  # type: ignore
        self.name = f"F({axes})"

    def forward(self, x: np.ndarray) -> np.ndarray:
        if any(x.shape[a] % 2 for a in self.axes):
            raise ValueError(f"Only even shapes are allowed for Fourrier transform, got {x.shape}")
        return np.fft.rfftn(x, axes=self.axes, norm="ortho")  # type: ignore

    def backward(self, y: np.ndarray) -> np.ndarray:
        return np.fft.irfftn(y, axes=self.axes, norm="ortho")  # type: ignore



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import typing as tp
from unittest.mock import patch
import numpy as np
from . import core


def test_powersystem_small() -> None:
    np.random.seed(12)
    dams = 2
    func = core.PowerSystem(num_dams=dams, num_years=0.2)
    x = [np.random.rand(func.dimension // dams) for _ in range(dams)]
    value = func.function(*x)
    np.testing.assert_almost_equal(value, 4266.8177479)


@patch(f"{__name__}.core.plt")
def test_make_plots(mock_plt: tp.Any) -> None:
    func = core.PowerSystem()
    func.losses = [0.1]
    func.make_plots("not_valid.png")
    assert mock_plt.clf.call_count == 1
    assert mock_plt.subplot.call_count == 4
    assert mock_plt.savefig.call_count == 1
